static char *sccsid = "@(#) samples/pubsub/amqspsda.c, pubsub, p000, p000-L000411          1.30  00/04/10 16:46:46";
/******************************************************************************/
/*                                                                            */
/* Module name: AMQSPSDA.C                                                    */
/*                                                                            */
/* Description: Sample C program for dumping the MQSeries Message Broker      */
/*              stream data.                                                  */
/*                                                                            */
/* Statement:   Licensed Materials - Property of IBM                          */
/*                                                                            */
/*              63H9336                                                       */
/*              (C) Copyright IBM Corp. 1998                                  */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/*                                                                            */
/*   AMQSPSDA is a sample C program for MQSeries Message Broker, it dumps     */
/*   the data from the brokers streams depending upon the arguments (see      */
/*   below) set. By subscribing using the MQREGO_PUBLISH_ON_REQUEST_ONLY      */
/*   option, the program will receive only retained messages when it issues   */
/*   the request update command. Note that unlike other MQSeries Message      */
/*   Broker programs this uses the MQSeries Administration Interface (MQAI)   */
/*   as MQSeries Message Broker administration messages are in PCF format.    */
/*                                                                            */
/* Program logic:                                                             */
/*                                                                            */
/*   check program arguments                                                  */
/*   MQCONN connect to default queue manager (or optionally supplied name)    */
/*   MQOPEN open broker control queue for output                              */
/*   create a reply queue if not supplied with one                            */
/*   MQOPEN open reply queue for input                                        */
/*   Dump relations                                                           */
/*   Dump streams supported                                                   */
/*   Dump information on publishers                                           */
/*   Dump information on subscribers                                          */
/*   Dump retained messages                                                   */
/*   if we created a reply queue                                              */
/*      MQCLOSE close reply queue with MQCO_DELETE_PURGE option               */
/*   else                                                                     */
/*      MQCLOSE close reply queue                                             */
/*   MQCLOSE close control queue                                              */
/*   MQDISC disconnect from queue manager                                     */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* AMQSPSDA has the following parameters:                                     */
/*                                                                            */
/* -m QMgrName (optional)                                                     */
/*    Default: the default queue manager will be used.                        */
/*                                                                            */
/* -q QName (optional) queue used to subscribe to.                            */
/*    Default: the program will attempt to create a permanent dynamic queue   */
/*    based on AMQSPSDA.PERMDYN.MODEL.QUEUE, it will be deleted at program    */
/*    termination.                                                            */
/*                                                                            */
/* -s StreamName (optional)                                                   */
/*    Default: all streams will be dumped.                                    */
/*                                                                            */
/* -l LogFileName (optional)                                                  */
/*    Default: output will be to screen.                                      */
/*                                                                            */
/* -t Topic (optional) used for publishers/subscribers subscriptions          */
/*    Default: * will be used as the topic.                                   */
/*                                                                            */
/* -p (optional) Dump information for publishers only.                        */
/*    Default: dump information for publishers and subscribers.               */
/*                                                                            */
/* -u (optional) Dump information for subscribers only.                       */
/*    Default: dump information for publishers and subscribers.               */
/*                                                                            */
/* -a (optional) Dump information for anonymous publishers/subscribers.       */
/*    Default: Dump information for non-anonymous publishers/subscribers.     */
/*                                                                            */
/* -r Topic (optional) Dump retained messages for Topic                       */
/*    Default: do not dump retained messages                                  */
/*                                                                            */
/******************************************************************************/


/******************************************************************************/
/*                                                                            */
/* Includes                                                                   */
/*                                                                            */
/******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <locale.h>
#include <time.h>

#include <cmqc.h>                                /* MQI                       */
#include <cmqcfc.h>                              /* MQAI                      */
#include <cmqbc.h>                               /* MQAI                      */
#include <cmqpsc.h>                              /* MQI PubSub                */

/******************************************************************************/
/* The msSleep macro needs some platform specific headers                     */
/******************************************************************************/
#if MQAT_DEFAULT == MQAT_WINDOWS_NT
   #include <windows.h>
#elif MQAT_DEFAULT == MQAT_OS2
   #define INCL_DOSPROCESS
   #include <os2.h>
#elif MQAT_DEFAULT == MQAT_UNIX
   #include <sys/types.h>
   #include <sys/time.h>
#endif

#if MQAT_DEFAULT == MQAT_UNIX
   #include <unistd.h>                           /* required for getopt       */
#endif

/******************************************************************************/
/*                                                                            */
/* Defines                                                                    */
/*                                                                            */
/******************************************************************************/

#define OK       0                               /* define OK as zero         */
#define NOT_OK   1                               /* define NOT_OK as one      */

#ifndef ULONG
   #define ULONG unsigned long
#endif

#ifndef BOOL
   #define BOOL ULONG
#endif

#ifndef TRUE
   #define TRUE  1
#endif

#ifndef FALSE
   #define FALSE 0
#endif

#define AMQSPSD_USAGE 88                         /* display program usage     */

#define DEFAULT_WAIT_INTERVAL     8000L          /* 8 seconds                 */
#define EXTENDED_WAIT_INTERVAL  300000L          /* 5 minutes                 */
#define WAIT_INTERVAL_SEGMENT    10000L          /* 10 seconds                */

#define CHARS_PER_LINE  16                       /* used to format a message  */

#define BUFFER_SIZE      256
#define BUFFER_SIZE_MAX  512

#define MAX_MSG_DUMPED_SIZE 256

#define ADMIN_STREAM  "SYSTEM.BROKER.ADMIN.STREAM"
#define MODEL_QUEUE   "AMQSPSD.PERMDYN.MODEL.QUEUE"

/******************************************************************************/
/*                                                                            */
/* Structures                                                                 */
/*                                                                            */
/******************************************************************************/
typedef struct StreamListEntry StreamListEntry, *PStreamListEntry;

struct StreamListEntry
{
   MQCHAR StreamName[MQ_Q_NAME_LENGTH+1];        /* stream name               */
   PStreamListEntry pNextSLE;                    /* ptr to next in entry      */
};

/*----------------------------------------------------------------------------*/
/* Parser_State : Possible states of the NameValueString parser               */
/*----------------------------------------------------------------------------*/
typedef enum
{
  OutOfToken,
  InToken,
  InQuotes,
  EmbeddedQuote,
  EndOfToken
} Parser_State;

/******************************************************************************/
/*                                                                            */
/* Global variables                                                           */
/*                                                                            */
/******************************************************************************/
MQCHAR QMgrName[MQ_Q_MGR_NAME_LENGTH+1]="";      /* default Q manager name    */
MQCHAR StreamName[MQ_Q_NAME_LENGTH+1]="*";       /* * = all streams dumped    */
MQCHAR QName[MQ_Q_NAME_LENGTH+1]="";             /* queue name                */
MQCHAR LogFileName[BUFFER_SIZE]="";              /* log file name             */
FILE *pLogFile;                                  /* ptr to log file           */
MQCHAR Topic[BUFFER_SIZE]="*";                   /* * = all topics dumped     */
MQCHAR RetainedTopic[BUFFER_SIZE]="";            /* retained topic dumped     */
BOOL PubOnly=FALSE;                              /* publishers only boolean   */
BOOL SubOnly=FALSE;                              /* subscribers only boolean  */
BOOL Anonymous=FALSE;                            /* dump ananoymous info      */
MQHCONN hCon;                                    /* handle to connection      */
MQHOBJ hCQObj;                                   /* handle to ctl Q object    */
MQHOBJ hRQObj;                                   /* handle to reply Q object  */
MQMD md={MQMD_DEFAULT};                          /* message descriptor        */
PStreamListEntry pStreamList=NULL;               /* stream list               */


/******************************************************************************/
/*                                                                            */
/* The UNIX getopt function is not available on the PC platforms, so we have  */
/* provided our own.                                                          */
/*                                                                            */
/******************************************************************************/
#if MQAT_DEFAULT == MQAT_WINDOWS_NT || MQAT_DEFAULT == MQAT_OS2
   int getopt(int argc, char *argv[], char *opts);
   int optind=1;                                 /* getopt index              */
   int optopt;                                   /* getopt option             */
   char *optarg;                                 /* getopt argument           */
#elif MQAT_DEFAULT == MQAT_UNIX
   extern int optind;                            /* getopt index              */
   extern char* optarg;                          /* getopt argument           */
#endif

/******************************************************************************/
/*                                                                            */
/* Function Prototypes                                                        */
/*                                                                            */
/******************************************************************************/
MQLONG GetArgs(int argc, char *argv[]);
MQLONG OpenLog(void);
MQLONG CreateReplyQueue(void);
MQLONG DumpRelations(void);
MQLONG DumpStreamsSupported(void);
MQLONG RegSubscriber(PMQCHAR Topic, PMQCHAR StreamName);
MQLONG DeregSubscriber(PMQCHAR Topic, PMQCHAR StreamName);
MQLONG ReqUpdate(PMQCHAR Topic, PMQCHAR StreamName);
MQLONG DumpStringData(MQLONG Identifier);
MQLONG DumpPubSubData(PMQCHAR SubStr);
MQLONG DumpPublishersOrSubscribers(PMQCHAR PubOrSub);
MQLONG DumpRetainedMessages(void);
MQLONG DumpRetainedMessagesData(void);
MQLONG DumpRetainedPCFMessageData(MQLONG msgLen, PMQBYTE pMsgBuf, MQMD MD);
MQLONG DumpRetainedRFHMessageData(MQLONG msgLen, PMQBYTE pMsgBuf, MQMD MD);
void DumpMessageData(MQLONG DataLength, PMQBYTE Buffer);
MQLONG BuildStreamList(void);
MQLONG FreeStreamList(void);
MQLONG DumpPCFIntegerItemBySelector(MQHBAG Bag, MQLONG Selector, char *Title);
MQLONG DumpPCFStringItemBySelector(MQHBAG Bag, MQLONG Selector, char *Title);
MQLONG DumpPCFIntegerItemByIndex(MQHBAG Bag, MQLONG ItemIndex, char *Title);
MQLONG DumpPCFStringItemByIndex(MQHBAG Bag, MQLONG ItemIndex, char *Title);
MQLONG SendBrokerCmd(PMQCHAR CmdStr,             /* Send broker command       */
                     PMQHOBJ phObj,
                     MQHBAG  hBag,
                     MQLONG  WaitTime);
MQLONG GetRFHTagValue(PMQCHAR *ppNameValueString,
                      MQLONG  *pStringLength,
                      PMQCHAR  pTagName,
                      PMQCHAR *ppTagValue,
                      ULONG   *pTagValueLength);
MQLONG GetNextToken(PMQCHAR *ppNameValueString,
                    PMQLONG  pRemainingLength,
                    PMQBYTE *ppToken,
                    ULONG   *pTokenLength);

/******************************************************************************/
/*                                                                            */
/* Macros                                                                     */
/*                                                                            */
/******************************************************************************/

/*----------------------------------------------------------------------------*/
/* Millisecond sleep                                                          */
/*----------------------------------------------------------------------------*/
#if MQAT_DEFAULT == MQAT_WINDOWS_NT
   #define msSleep(time)                                                       \
      Sleep(time)
#elif MQAT_DEFAULT == MQAT_OS2
   #define msSleep(time)                                                       \
      DosSleep(time)
#elif MQAT_DEFAULT == MQAT_UNIX
   #define msSleep(time)                                                       \
   {                                                                           \
      struct timeval tval;                                                     \
                                                                               \
      tval.tv_sec  = time / 1000;                                              \
      tval.tv_usec = (time % 1000) * 1000;                                     \
                                                                               \
      select(0, NULL, NULL, NULL, &tval);                                      \
   }
#endif


/*----------------------------------------------------------------------------*/
/* Checks the comp code, if not MQCC_OK prints a message and sets return code */
/*----------------------------------------------------------------------------*/
#define mqCheck(func, compCode, reason)                                        \
{                                                                              \
   if (compCode != MQCC_OK)                                                    \
   {                                                                           \
      fprintf(pLogFile, func " failed with reason=%ld %lXX\n",                 \
              reason, reason);                                                 \
      fprintf(pLogFile, "preceding line %d\n", __LINE__);                      \
      if (rc == OK)                            /* don't overwrite a bad rc  */ \
         rc = compCode;                                                        \
   } /* endif */                                                               \
}


#define BlankPadName(BPadName, Name, NameLen)                                  \
{                                                                              \
   MQLONG strLen=strlen(Name);                                                 \
                                                                               \
   memcpy(BPadName, Name, strLen);                                             \
   if (strLen < NameLen)                                                       \
      memset(BPadName + strLen, ' ', NameLen - strLen);                        \
   BPadName[NameLen] = '\0';                   /* string term for strcat    */ \
}

/******************************************************************************/
/* Define the min (minimum) macro.                                            */
/******************************************************************************/
#ifndef min
   #define min(a,b)   (((a) < (b)) ? (a) : (b))
#endif

/******************************************************************************/
/*                                                                            */
/* Function: Main                                                             */
/* ==============                                                             */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* int argc                       - number of arguments supplied              */
/* char *argv[]                   - program arguments                         */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
int main(int argc, char *argv[])
{
   MQLONG rc=OK;                                 /* return code               */
   MQLONG count;                                 /* used for loop counting    */
   MQHOBJ hQMObj;                                /* handle to QMgr object     */
   MQOD QMOD={MQOD_DEFAULT};                     /* queue manager object desc */
   MQOD CQOD={MQOD_DEFAULT};                     /* control queue object desc */
   MQOD RQOD={MQOD_DEFAULT};                     /* reply queue object desc   */
   MQLONG options;                               /* options                   */
   MQLONG reason;                                /* reason code               */
   MQLONG connReason;                            /* MQCONN reason code        */
   MQLONG compCode;                              /* completion code           */
   MQLONG openCQCompCode=MQCC_FAILED;            /* MQOPEN control queue CC   */
   MQLONG openRQCompCode=MQCC_FAILED;            /* MQOPEN reply queue CC     */
   BOOL replyQueueCreated=FALSE;                 /* reply queue created bool  */
   MQLONG Selector[1];                           /* Selector array for MQINQ  */

   /***************************************************************************/
   /* First check we have been given correct arguments                        */
   /***************************************************************************/
   rc = GetArgs(argc, argv);

   if (rc != OK)
   {
      if (rc == AMQSPSD_USAGE)
         exit(OK);
      else
         exit((int) rc);                         /* no point in going on      */
   } /* endif */

   /***************************************************************************/
   /* Open the log                                                            */
   /***************************************************************************/
   if (rc == OK)
      rc = OpenLog();                            /* open the log              */

   /***************************************************************************/
   /* Connect to queue manager                                                */
   /***************************************************************************/
   if (rc == OK)
   {
      MQCONN(QMgrName, &hCon, &compCode, &connReason);
      mqCheck("MQCONN", compCode, connReason);
      if (compCode == MQCC_FAILED)
      {
         if (LogFileName[0] != '\0')             /* close the log file if we  */
            fclose(pLogFile);                    /* have been given a name    */
         exit((int) connReason);                 /* no point in going on      */
      } /* endif */
   } /* endif */

   /***************************************************************************/
   /* We need to know the name of the queue manager for topic subscription,   */
   /* if one was not supplied as a parameter we must 'inquire' it from the    */
   /* queue manager.                                                          */
   /***************************************************************************/
   if (rc == OK && QMgrName[0] == '\0')
   {
      options = MQOO_INQUIRE;
      QMOD.ObjectType = MQOT_Q_MGR;
      MQOPEN(hCon, &QMOD, options, &hQMObj, &compCode, &reason);
      mqCheck("MQOPEN", compCode, reason);

      if (rc == OK)
      {
         Selector[0] = MQCA_Q_MGR_NAME;       /* set selector to QMgrName  */
         MQINQ(hCon, hQMObj,
               1L, Selector,                  /* selector count, array     */
               0L, NULL,                      /* int attr count, array     */
               MQ_Q_MGR_NAME_LENGTH,          /* char attr length          */
               QMgrName,                      /* char attr array           */
               &compCode, &reason);
         mqCheck("MQINQ", compCode, reason);

         /*-------------------------------------------------------------------*/
         /* Issue the MQCLOSE reqardless                                      */
         /*-------------------------------------------------------------------*/
         MQCLOSE(hCon, &hQMObj, MQCO_NONE, &compCode, &reason);
         mqCheck("MQCLOSE", compCode, reason);
      } /* endif */
   } /* endif */

   /***************************************************************************/
   /* Open the control queue for output                                       */
   /***************************************************************************/
   if (rc == OK)
   {
      strcpy(CQOD.ObjectName, "SYSTEM.BROKER.CONTROL.QUEUE");
      options = MQOO_OUTPUT + MQOO_FAIL_IF_QUIESCING;
      MQOPEN(hCon, &CQOD, options, &hCQObj, &openCQCompCode, &reason);
      mqCheck("MQOPEN", openCQCompCode, reason);
   } /* endif */

   /***************************************************************************/
   /* If we haven't been given a queue, create a permanent dynamic one        */
   /***************************************************************************/
   if (rc == OK && QName[0] == '\0')
   {
      rc = CreateReplyQueue();
      if (rc == OK)
         replyQueueCreated = TRUE;
   } /* endif */

   /***************************************************************************/
   /* Open the reply queue for input                                          */
   /***************************************************************************/
   if (rc == OK)
   {
      strcpy(RQOD.ObjectName, QName);
      options = MQOO_INPUT_SHARED + MQOO_FAIL_IF_QUIESCING;
      MQOPEN(hCon, &RQOD, options, &hRQObj, &openRQCompCode, &reason);
      mqCheck("MQOPEN", openRQCompCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Set the message descriptor                                              */
   /***************************************************************************/
   if (rc == OK)
   {
      memcpy(&md.Format, MQFMT_PCF, sizeof(md.Format));
      md.MsgType = MQMT_REQUEST;
      memcpy(md.ReplyToQ, QName, MQ_Q_NAME_LENGTH);
      memcpy(md.ReplyToQMgr, QMgrName, MQ_Q_MGR_NAME_LENGTH);
      md.Expiry = 3000;                          /* 300 second expiry         */
      md.Report = MQRO_DISCARD_MSG;
   } /* endif */

   /***************************************************************************/
   /* All the setup is now done, start dumping information                    */
   /***************************************************************************/
   /***************************************************************************/
   /* Dump relations                                                          */
   /***************************************************************************/
   if (rc == OK)
      rc = DumpRelations();

   /***************************************************************************/
   /* Dump streams supported                                                  */
   /***************************************************************************/
   if (rc == OK)
   {
      if (strcmp(StreamName, "*") == 0)
      {
         rc = DumpStreamsSupported();
      }
      else
      {
         /*-------------------------------------------------------------------*/
         /* We have been given a StreamName, so build a list with it only     */
         /*-------------------------------------------------------------------*/
         PStreamListEntry pNewSLE;               /* -> new stream list entry  */

         pNewSLE = malloc(sizeof(StreamListEntry));

         if (pNewSLE != NULL)
         {
            strcpy(pNewSLE->StreamName, StreamName);
            pNewSLE->pNextSLE = pStreamList;
            pStreamList = pNewSLE;
         }
         else
         {
            fprintf(pLogFile, "Error malloc failed !\n");
            rc = NOT_OK;
         } /* endif */
      } /* endif */
   } /* endif */

   /***************************************************************************/
   /* Dump information on publishers                                          */
   /***************************************************************************/
   if (rc == OK && !SubOnly)
      rc = DumpPublishersOrSubscribers("Publishers");

   /***************************************************************************/
   /* Dump information on subscribers                                         */
   /***************************************************************************/
   if (rc == OK && !PubOnly)
      rc = DumpPublishersOrSubscribers("Subscribers");

   /***************************************************************************/
   /* Dump retained messages                                                  */
   /***************************************************************************/
   if (rc == OK && RetainedTopic[0] != '\0')
      rc = DumpRetainedMessages();

   /***************************************************************************/
   /* Finished so start cleaning up                                           */
   /***************************************************************************/
   /***************************************************************************/
   /* Free stream list                                                        */
   /***************************************************************************/
   if(pStreamList != NULL)
      rc = FreeStreamList();

   /***************************************************************************/
   /* If we created a reply queue, delete it                                  */
   /***************************************************************************/
   if (replyQueueCreated)
   {
      MQCLOSE(hCon, &hRQObj, MQCO_DELETE_PURGE, &compCode, &reason);

      /************************************************************************/
      /* If there are uncommitted get or put requests outstanding, we will    */
      /* have reason equal to MQRC_Q_NOT_EMPTY, in which case we will keep    */
      /* trying for up to 5 minutes to get rid of the queue.                  */
      /************************************************************************/
      if (reason == MQRC_Q_NOT_EMPTY)
      {
         for (count=0; reason == MQRC_Q_NOT_EMPTY && count < 30; count++)
         {
            msSleep(10000);                      /* sleep 10 seconds          */
            MQCLOSE(hCon, &hRQObj, MQCO_DELETE_PURGE, &compCode, &reason);
            /*----------------------------------------------------------------*/
            /* If we still haven't managed it yet, inform user.               */
            /*----------------------------------------------------------------*/
            if (reason == MQRC_Q_NOT_EMPTY && count == 1)
               fprintf(pLogFile, "Trying to delete reply queue, will keep trying "
                                 "for up to 5 minutes");
            else if (reason == MQRC_Q_NOT_EMPTY && count > 1)
               fprintf(pLogFile, ".");
         } /* endfor */

         if (count > 1)
            fprintf(pLogFile, "\n");
      } /* endif */

      mqCheck("MQCLOSE", compCode, reason);
   }
   else if (openRQCompCode != MQCC_FAILED)
   {
      MQCLOSE(hCon, &hRQObj, MQCO_NONE, &compCode, &reason);
      mqCheck("MQCLOSE", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Close queues if opened                                                  */
   /***************************************************************************/
   if (openCQCompCode != MQCC_FAILED)
   {
      MQCLOSE(hCon, &hCQObj, MQCO_NONE, &compCode, &reason);
      mqCheck("MQCLOSE", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Disconnect from queue manager if not already connected                  */
   /***************************************************************************/
   if (connReason != MQRC_ALREADY_CONNECTED)
   {
      MQDISC(&hCon, &compCode, &reason);
      mqCheck("MQDISC", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Close the log file if we have been given a name                         */
   /***************************************************************************/
   if (LogFileName[0] != '\0')
      fclose(pLogFile);

   return ((int) rc);
}                                                /* ------ END OF MAIN ------ */


/******************************************************************************/
/*                                                                            */
/* Function: GetArgs                                                          */
/*                                                                            */
/* Description: Get the program arguments.                                    */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Check all the program arguments are valid and copy them into appropriate   */
/* variables.                                                                 */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* int argc                       - number of arguments supplied              */
/* char *argv[]                   - program arguments                         */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
MQLONG GetArgs(int argc, char *argv[])
{
   MQLONG rc=OK;                                 /* return code               */
   int o;                                        /* option character          */
   int numM=0;                                   /* number of 'm' options     */
   int numQ=0;                                   /* number of 'q' options     */
   int numS=0;                                   /* number of 's' options     */
   int numL=0;                                   /* number of 'l' options     */
   int numT=0;                                   /* number of 't' options     */
   int numP=0;                                   /* number of 'p' options     */
   int numU=0;                                   /* number of 'u' options     */
   int numA=0;                                   /* number of 'a' options     */
   int numR=0;                                   /* number of 'r' options     */

   /***************************************************************************/
   /* If there is a single argument which is just a question mark, then just  */
   /* issue the command's usage message                                       */
   /***************************************************************************/
   if ((argc == 2) && (strcmp(argv[1], "?") == 0))
      rc = AMQSPSD_USAGE;

   if (rc == OK)
   {
      while ((o = getopt(argc, argv, "m:q:s:l:t:puar:")) != EOF)
      {
         switch (o)
         {
            case 'm':
               numM++;
               if (strlen(optarg) <= MQ_Q_MGR_NAME_LENGTH)
               {
                  strcpy(QMgrName, optarg);
               }
               else
               {
                  fprintf(stderr, "[-m] QMgrName too long, max length %d\n", MQ_Q_MGR_NAME_LENGTH);
                  rc = 99;
               } /* endif */
               break;

            case 'q':
               numQ++;
               if (strlen(optarg) <= MQ_Q_NAME_LENGTH)
               {
                  strcpy(QName, optarg);
               }
               else
               {
                  fprintf(stderr, "[-q] QName too long, max length %d\n", MQ_Q_NAME_LENGTH);
                  rc = 99;
               } /* endif */
               break;

            case 's':
               numS++;
               if (strlen(optarg) <= MQ_Q_NAME_LENGTH)
               {
                  strcpy(StreamName, optarg);
               }
               else
               {
                  fprintf(stderr, "[-s] StreamName too long, max length %d\n", MQ_Q_NAME_LENGTH);
                  rc = 99;
               } /* endif */
               break;

            case 'l':
               numL++;
               if (strlen(optarg) < BUFFER_SIZE)
               {
                  strcpy(LogFileName, optarg);
               }
               else
               {
                  fprintf(stderr, "[-l] LogFileName too long, max length %d\n", BUFFER_SIZE -1);
                  rc = 99;
               } /* endif */
               break;

            case 't':
               numT++;
               if (strlen(optarg) < BUFFER_SIZE)
               {
                  strcpy(Topic, optarg);
               }
               else
               {
                  fprintf(stderr, "[-t] Topic too long, max length %d\n", BUFFER_SIZE -1);
                  rc = 99;
               } /* endif */
               break;

            case 'p':
               numP++;
               PubOnly = TRUE;
               break;

            case 'u':
               numU++;
               SubOnly = TRUE;
               break;

            case 'a':
               numA++;
               Anonymous = TRUE;
               break;

            case 'r':
               numR++;
               if (strlen(optarg) < BUFFER_SIZE)
               {
                  strcpy(RetainedTopic, optarg);
               }
               else
               {
                  fprintf(stderr, "[-r] RetainedTopic too long, max length %d\n", BUFFER_SIZE -1);
                  rc = 99;
               } /* endif */
               break;

            /*----------------------------------------------------------------*/
            /* If an unknown option is supplied or an option with a ':' does  */
            /* not have an argument, then '?' is returned from getopt.        */
            /*----------------------------------------------------------------*/
            case '?':
               rc = 99;
               break;
         } /* endswitch */
      } /* endwhile */

      /************************************************************************/
      /* Make sure we were not given an option the wrong number of times      */
      /************************************************************************/
      if (numM > 1 || numQ > 1 || numS > 1 || numL > 1 || numT > 1 || numP > 1 ||
          numU > 1 || numA > 1 || numR > 1)
         rc = 99;

      /************************************************************************/
      /* Make sure there are no unused arguments                              */
      /************************************************************************/
      if (optind != argc)
         rc = 99;

   } /* endif */

   if (rc != OK)
   {
      fprintf(stderr, "Usage: %s [ -m <QMgrName> ]\n", argv[0]);
      fprintf(stderr, "               [ -q <QName> ]\n");
      fprintf(stderr, "               [ -s <StreamName> ]\n");
      fprintf(stderr, "               [ -l <LogFileName> ]\n");
      fprintf(stderr, "               [ -t <Topic> ]\n");
      fprintf(stderr, "               [ -p ] Dump information for publishers only\n");
      fprintf(stderr, "               [ -u ] Dump information for subscribers only\n");
      fprintf(stderr, "               [ -a ] Dump information for anonymous publishers/subscribers\n");
      fprintf(stderr, "               [ -r <RetainedTopic> ] Dump retained messages\n");
   } /* endif */

   return rc;
}


/******************************************************************************/
/*                                                                            */
/* Function: OpenLog                                                          */
/*                                                                            */
/* Description: Open the log file.                                            */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Open the log file if we have been given a name, else use stdout.           */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* None                                                                       */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
MQLONG OpenLog(void)
{
   MQLONG rc=OK;                                 /* return code               */
   time_t curAbsTime;                            /* current absolute time     */
   struct tm *pTimeStrt;                         /* ptr to time structure     */
   char *pTimeStr;                               /* ptr to time string        */

   /***************************************************************************/
   /* If we have been given a log file name open file else use stdout         */
   /***************************************************************************/
   if (LogFileName[0] != '\0')
   {
      pLogFile = fopen(LogFileName, "w");        /* create/truncate to zero   */
                                                 /* text file for write       */
      if (pLogFile != NULL)
      {
         /*-------------------------------------------------------------------*/
         /* Reopen text file for writing at EOF                               */
         /*-------------------------------------------------------------------*/
         pLogFile = freopen(LogFileName, "a", pLogFile);
      } /* endif */

      if (pLogFile == NULL)
      {
         fprintf(stderr, "Could not open LogFileName: %s\n", LogFileName);
         rc = NOT_OK;
      } /* endif */
   }
   else
   {
      pLogFile = (FILE *) stdout;
   } /* endif */

   if (rc == OK)
   {
      time(&curAbsTime);                         /* get absolute start time   */
      pTimeStrt = localtime(&curAbsTime);        /* convert to struct tm form */
      pTimeStr = asctime(pTimeStrt);             /* convert to string         */

      fprintf(pLogFile, "\nMQSeries Message Broker Dumper\nStart time: %s", pTimeStr);
   } /* endif */

   return rc;
}


/******************************************************************************/
/*                                                                            */
/* Function: CreateReplyQueue                                                 */
/*                                                                            */
/* Description: Create a permanent dynamic queue as the reply queue           */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Creates a permanent dynamic queue as the reply queue                       */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* None                                                                       */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
MQLONG CreateReplyQueue(void)
{
   MQLONG rc=OK;                                 /* return code               */
   MQHOBJ hRQObj;                                /* handle to reply Q object  */
   MQOD od={MQOD_DEFAULT};                       /* object descriptor         */
   MQLONG options;                               /* options                   */
   MQLONG openRQCompCode=MQCC_FAILED;            /* MQOPEN completion code    */
   MQLONG reason;                                /* reason code               */
   MQLONG compCode;                              /* completion code           */
   MQCHAR buffer[BUFFER_SIZE]="";                /* buffer                    */

   if (rc == OK)
   {
      /************************************************************************/
      /* Include the calling program name in the name of the new queue        */
      /************************************************************************/
      strcpy(od.ObjectName, MODEL_QUEUE);
      strcpy(od.DynamicQName, "amqspsd.*");

      /************************************************************************/
      /* Open the queue                                                       */
      /************************************************************************/
      options = MQOO_INPUT_SHARED + MQOO_FAIL_IF_QUIESCING;
      MQOPEN(hCon, &od, options, &hRQObj, &openRQCompCode, &reason);
      mqCheck("MQOPEN", openRQCompCode, reason);
   } /* endif */

   /***************************************************************************/
   /* If the model queue does not exist, print message                        */
   /***************************************************************************/
   if(reason == MQRC_UNKNOWN_OBJECT_NAME)
   {
      fprintf(pLogFile, "\n%s does not exist!\n"
              "Restart the program and supply a QName or create the model queue:\n\n"
              "runmqsc QMgrName < amqspsda.tst\n\nor\n\n"
              "def qmodel(%s) replace +\n"
              "    descr('Message Broker Dumper model queue') +\n"
              "    deftype(permdyn) +\n"
              "    share +\n"
              "    defsopt(shared)\n", MODEL_QUEUE, MODEL_QUEUE);
   } /* endif */

   if (rc == OK)
   {
      memcpy(QName, od.ObjectName, MQ_Q_NAME_LENGTH);
      QName[MQ_Q_NAME_LENGTH] = '\0';
   } /* endif */

   /***************************************************************************/
   /* Close queue if opened                                                   */
   /***************************************************************************/
   if (openRQCompCode != MQCC_FAILED)
   {
      MQCLOSE(hCon, &hRQObj, MQCO_NONE, &compCode, &reason);
      mqCheck("MQCLOSE", compCode, reason);
   } /* endif */

   return(rc);
}


/******************************************************************************/
/*                                                                            */
/* Function: DumpRelations                                                    */
/*                                                                            */
/* Description: Dump the broker relations                                     */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Dumps the brokers parent and children                                      */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* None                                                                       */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
MQLONG DumpRelations(void)
{
   MQLONG rc=OK;                                 /* return code               */
   MQLONG rc2=OK;                                /* return code 2             */
   MQCHAR topic[BUFFER_SIZE];                    /* topic                     */
   MQCHAR streamName[]=ADMIN_STREAM;             /* stream name               */
   MQCHAR BPadQMgrName[MQ_Q_MGR_NAME_LENGTH+1];  /* queue manager name        */
   BOOL registered=FALSE;                        /* registered boolean        */

   /***************************************************************************/
   /* The QMgrName in admin topics needs to be blank paded to full length     */
   /***************************************************************************/
   BlankPadName(BPadQMgrName, QMgrName, MQ_Q_MGR_NAME_LENGTH);

   strcpy(topic, "MQ/");
   strcat(topic, BPadQMgrName);
   strcat(topic, "/Parent");

   if (rc == OK)
      rc = RegSubscriber(topic, streamName);

   /***************************************************************************/
   /* This is our first interaction with the broker, if this fails it could   */
   /* be due to the broker not being up.                                      */
   /***************************************************************************/
   if (rc != OK)
   {
      fprintf(pLogFile, "\n---> Failed to talk to the broker, is it running ? <---\n\n");

      /************************************************************************/
      /* Even though it appears the broker is not up, just in case it's not   */
      /* very well, we will deregister.                                       */
      /************************************************************************/
      rc2 = DeregSubscriber(topic, streamName);
   }
   else
   {
      registered = TRUE;
      fprintf(pLogFile, "\nBroker Relations\n----------------\nQMgrName:\n   %s\n"
                        "Parent:\n", QMgrName);
      rc = ReqUpdate(topic, streamName);

      if (rc == MQRCCF_NO_RETAINED_MSG)          /* this rc is not an error   */
      {
         rc = OK;
         fprintf(pLogFile, "   None\n");
      }
      else if (rc == OK)
      {
         rc = DumpStringData(MQCACF_PARENT_Q_MGR_NAME);
      } /* endif */
   } /* endif */

   if (rc == OK && registered)
      rc = DeregSubscriber(topic, streamName);
   else if (registered)
      rc2 = DeregSubscriber(topic, streamName);

   if (rc == OK)
   {
      strcpy(topic, "MQ/");
      strcat(topic, BPadQMgrName);
      strcat(topic, "/Children");
   } /* endif */

   registered = FALSE;                           /* reset flag                */

   if (rc == OK)
      rc = RegSubscriber(topic, streamName);

   if (rc == OK)
   {
      registered = TRUE;
      fprintf(pLogFile, "Children:\n");
      rc = ReqUpdate(topic, streamName);

      if (rc == MQRCCF_NO_RETAINED_MSG)          /* this rc is not an error   */
      {
         rc = OK;
         fprintf(pLogFile, "   None\n");
      }
      else if (rc == OK)
      {
         rc = DumpStringData(MQCACF_CHILD_Q_MGR_NAME);
      } /* endif */
   } /* endif */

   if (rc == OK && registered)
      rc = DeregSubscriber(topic, streamName);
   else if (registered)
      rc2 = DeregSubscriber(topic, streamName);

   fflush(pLogFile);                             /* force the output          */

   if (rc == OK)
      rc = rc2;

   return(rc);
}


/******************************************************************************/
/*                                                                            */
/* Function: DumpStreamsSupported                                             */
/*                                                                            */
/* Description: Dump the streams supported                                    */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Dumps the streams supported by this broker                                 */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* None                                                                       */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
MQLONG DumpStreamsSupported(void)
{
   MQLONG rc=OK;                                 /* return code               */
   MQLONG rc2=OK;                                /* return code 2             */
   MQCHAR topic[BUFFER_SIZE];                    /* topic                     */
   MQCHAR streamName[]=ADMIN_STREAM;             /* stream name               */
   MQCHAR BPadQMgrName[MQ_Q_MGR_NAME_LENGTH+1];  /* queue manager name        */
   BOOL registered=FALSE;                        /* registered boolean        */

   /***************************************************************************/
   /* The QMgrName in admin topics needs to be blank paded to full length     */
   /***************************************************************************/
   BlankPadName(BPadQMgrName, QMgrName, MQ_Q_MGR_NAME_LENGTH);

   strcpy(topic, "MQ/");
   strcat(topic, BPadQMgrName);
   strcat(topic, "/StreamSupport");

   if (rc == OK)
      rc = RegSubscriber(topic, streamName);

   if (rc == OK)
   {
      registered = TRUE;
      fprintf(pLogFile, "\nStreams supported\n-----------------\n");
      rc = ReqUpdate(topic, streamName);

      if (rc == MQRCCF_NO_RETAINED_MSG)          /* this rc is not an error   */
      {
         rc = OK;
         fprintf(pLogFile, "None\n");
      }
      else if (rc == OK)
      {
         rc = BuildStreamList();
      } /* endif */
   } /* endif */

   if (rc == OK && registered)
      rc = DeregSubscriber(topic, streamName);
   else if (registered)
      rc2 = DeregSubscriber(topic, streamName);

   fflush(pLogFile);                             /* force the output          */

   if (rc == OK)
      rc = rc2;

   return(rc);
}


/******************************************************************************/
/*                                                                            */
/* Function: DumpPublishersOrSubscribers                                      */
/*                                                                            */
/* Description: Dump the publishers or subscribers and their details          */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Dumps the publishers or subscribers and their details for each stream      */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* PMQCHAR PubOrSub               - 'Publishers' or 'Subscribers'             */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
MQLONG DumpPublishersOrSubscribers(PMQCHAR PubOrSub)
{
   MQLONG rc=OK;                                 /* return code               */
   MQLONG rc2=OK;                                /* return code 2             */
   MQLONG count;                                 /* used for loop counting    */
   MQLONG strLen;                                /* string length             */
   MQCHAR topic[BUFFER_SIZE_MAX];                /* topic                     */
   MQCHAR topic2[BUFFER_SIZE];                   /* topic 2                   */
   MQCHAR streamName[MQ_Q_NAME_LENGTH+1];        /* stream name               */
   MQCHAR BPadQMgrName[MQ_Q_MGR_NAME_LENGTH+1];  /* queue manager name        */
   BOOL registered=FALSE;                        /* registered boolean        */
   PStreamListEntry pSLE;                        /* -> stream list entry      */

   fprintf(pLogFile, "\n%s\n", PubOrSub);
   strLen = strlen(PubOrSub);                    /* underline heading         */
   for (count=0; count < strLen; count++)
      fprintf(pLogFile, "-");
   fprintf(pLogFile, "\n");

   /***************************************************************************/
   /* The QMgrName in admin topics needs to be blank paded to full length     */
   /***************************************************************************/
   BlankPadName(BPadQMgrName, QMgrName, MQ_Q_MGR_NAME_LENGTH);

   /***************************************************************************/
   /* Build up the topic string. The global Topic string may well be "*", we  */
   /* dont want to dump this out, only the real topics published or sub-      */
   /* to, hence save the topic without the global Topic.                      */
   /***************************************************************************/
   if (Anonymous)
      strcpy(topic, "MQ/SA/");
   else
      strcpy(topic, "MQ/S/");
   strcat(topic, BPadQMgrName);
   strcat(topic, "/");
   strcat(topic, PubOrSub);
   if (Anonymous)
      strcat(topic, "/AllIdentities/");
   else
      strcat(topic, "/Identities/");
   strcpy(topic2, topic);                        /* save without Topic        */
   strcat(topic, Topic);                         /* add the global Topic      */

   /***************************************************************************/
   /* For each stream dump the publishers or subscribers and their details    */
   /***************************************************************************/
   if (rc == OK)
   {
      for (pSLE=pStreamList; pSLE != NULL && rc == OK; pSLE=pSLE->pNextSLE)
      {
         strcpy(streamName, pSLE->StreamName);
         fprintf(pLogFile, "StreamName: %s\n", streamName);

         if (rc == OK)
            rc = RegSubscriber(topic, streamName);

         if (rc == OK)
         {
            registered = TRUE;
            rc = ReqUpdate(topic, streamName);

            if (rc == MQRCCF_NO_RETAINED_MSG)    /* this rc is not an error   */
            {
               rc = OK;
               fprintf(pLogFile, "   None\n");
            }
            else if (rc == OK)
            {
               rc = DumpPubSubData(topic2);
            } /* endif */
         } /* endif */

         if (rc == OK && registered)
         {
            rc = DeregSubscriber(topic, streamName);
            registered = FALSE;                  /* reset flag                */
         }
         else if (registered)
         {
            rc2 = DeregSubscriber(topic, streamName);
            registered = FALSE;                  /* reset flag                */
         } /* endif */
      } /* endfor */
   } /* endif */

   fflush(pLogFile);                             /* force the output          */

   if (rc == OK)                                 /* return a bad rc if either */
      rc = rc2;                                  /* rc or rc2 is bad          */

   return(rc);
}


/******************************************************************************/
/*                                                                            */
/* Function: DumpRetainedMessages                                             */
/*                                                                            */
/* Description: Dump retained messages                                        */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Dump retained messages for the supplied topic                              */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* None                                                                       */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
MQLONG DumpRetainedMessages(void)
{
   MQLONG rc=OK;                                 /* return code               */
   MQLONG rc2=OK;                                /* return code 2             */
   MQCHAR streamName[MQ_Q_NAME_LENGTH+1];        /* stream name               */
   BOOL registered=FALSE;                        /* registered boolean        */
   PStreamListEntry pSLE;                        /* -> stream list entry      */

   fprintf(pLogFile, "\nRetained messages\n-----------------\n");

   /***************************************************************************/
   /* For each stream dump the retained messages for supplied topic           */
   /***************************************************************************/
   if (rc == OK)
   {
      for (pSLE=pStreamList; pSLE != NULL && rc == OK; pSLE=pSLE->pNextSLE)
      {
         /*-------------------------------------------------------------------*/
         /* All the messages on the ADMIN_STREAM are retained, we have        */
         /* already dumped everything we are interested in on this stream, so */
         /* only dump retained messages for other streams.                    */
         /*-------------------------------------------------------------------*/
         if (memcmp(pSLE->StreamName, ADMIN_STREAM, strlen(ADMIN_STREAM)) != 0)
         {
            strcpy(streamName, pSLE->StreamName);
            fprintf(pLogFile, "StreamName: %s\n", streamName);

            if (rc == OK)
               rc = RegSubscriber(RetainedTopic, streamName);

            if (rc == OK)
            {
               registered = TRUE;
               rc = ReqUpdate(RetainedTopic, streamName);

               if (rc == MQRCCF_NO_RETAINED_MSG) /* this rc is not an error   */
               {
                  rc = OK;
                  fprintf(pLogFile, " None\n");
               }
               else if (rc == OK)
               {
                  rc = DumpRetainedMessagesData();
               } /* endif */
            } /* endif */

            if (rc == OK && registered)
            {
               rc = DeregSubscriber(RetainedTopic, streamName);
               registered = FALSE;               /* reset flag                */
            }
            else if (registered)
            {
               rc2 = DeregSubscriber(RetainedTopic, streamName);
               registered = FALSE;               /* reset flag                */
            } /* endif */
         } /* endif */
      } /* endfor */
   } /* endif */

   fflush(pLogFile);                             /* force the output          */

   if (rc == OK)                                 /* return a bad rc if either */
      rc = rc2;                                  /* rc or rc2 is bad          */

   return(rc);
}


/******************************************************************************/
/*                                                                            */
/* Function: RegSubscriber                                                    */
/*                                                                            */
/* Description: Register Subscriber                                           */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Implements the MQCMD_REGISTER_SUBSCRIBER command message.                  */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* PMQCHAR Topic                  - topic                                     */
/* PMQCHAR StreamName             - stream name                               */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
MQLONG RegSubscriber(PMQCHAR Topic, PMQCHAR StreamName)
{
   MQLONG rc=OK;                                 /* return code               */
   MQHBAG bag=MQHB_UNUSABLE_HBAG;                /* MQ bag                    */
   MQLONG reason;                                /* reason code               */
   MQLONG compCode;                              /* completion code           */
   MQLONG regOptions;                            /* registration options      */

   /***************************************************************************/
   /* Create bag which we will use to build up the request to the broker      */
   /***************************************************************************/
   if (rc == OK)
   {
      mqCreateBag(MQCBO_COMMAND_BAG, &bag, &compCode, &reason);
      mqCheck("mqCreateBag", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Set command in bag                                                      */
   /***************************************************************************/
   if (rc == OK)
   {
      mqSetInteger(bag, MQIASY_COMMAND, 0, MQCMD_REGISTER_SUBSCRIBER,
                   &compCode, &reason);
      mqCheck("mqSetInteger", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Add Topic to bag                                                        */
   /***************************************************************************/
   if (rc == OK)
   {
      mqAddString(bag, MQCACF_TOPIC, MQBL_NULL_TERMINATED, Topic,
                  &compCode, &reason);
      mqCheck("mqAddString", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Add RegistrationOptions to bag                                          */
   /***************************************************************************/
   if (rc == OK)
   {
      /*----------------------------------------------------------------------*/
      /* MQREGO_ANONYMOUS is used so that the broker will not publish our     */
      /* identity. MQREGO_PUBLISH_ON_REQUEST_ONLY is used so the we will only */
      /* be sent retained publications, and only when we issue a request      */
      /* update command.                                                      */
      /*----------------------------------------------------------------------*/
      regOptions = MQREGO_ANONYMOUS | MQREGO_PUBLISH_ON_REQUEST_ONLY;
      mqAddInteger(bag, MQIACF_REGISTRATION_OPTIONS, regOptions,
                   &compCode, &reason);
      mqCheck("mqAddInteger", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Add StreamName to bag                                                   */
   /***************************************************************************/
   if (rc == OK)
   {
      mqAddString(bag, MQCACF_STREAM_NAME, MQBL_NULL_TERMINATED, StreamName,
                  &compCode, &reason);
      mqCheck("mqAddString", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Add QMgrName to bag                                                     */
   /***************************************************************************/
   if (rc == OK)
   {
      mqAddString(bag, MQCA_Q_MGR_NAME, MQBL_NULL_TERMINATED, QMgrName,
                  &compCode, &reason);
      mqCheck("mqAddString", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Add QName to bag                                                        */
   /***************************************************************************/
   if (rc == OK)
   {
      mqAddString(bag, MQCA_Q_NAME, MQBL_NULL_TERMINATED, QName,
                  &compCode, &reason);
      mqCheck("mqAddString", compCode, reason);
   } /* endif */

   if (rc == OK)
      rc = SendBrokerCmd("Register Subscriber", &hCQObj, bag, DEFAULT_WAIT_INTERVAL);

   /***************************************************************************/
   /* Delete the bag if we managed to create one                              */
   /***************************************************************************/
   if (bag != MQHB_UNUSABLE_HBAG)
   {
      mqDeleteBag(&bag, &compCode, &reason);
      mqCheck("mqDeleteBag", compCode, reason);
   } /* endif */

   return(rc);
}


/******************************************************************************/
/*                                                                            */
/* Function: DeregSubscriber                                                  */
/*                                                                            */
/* Description: Deregister Subscriber                                         */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Implements the MQCMD_DEREGISTER_SUBSCRIBER command message.                */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* PMQCHAR Topic                  - topic                                     */
/* PMQCHAR StreamName             - stream name                               */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
MQLONG DeregSubscriber(PMQCHAR Topic, PMQCHAR StreamName)
{
   MQLONG rc=OK;                                 /* return code               */
   MQHBAG bag=MQHB_UNUSABLE_HBAG;                /* MQ bag                    */
   MQLONG reason;                                /* reason code               */
   MQLONG compCode;                              /* completion code           */
   MQLONG regOptions=0;                          /* registration options      */

   /***************************************************************************/
   /* Create bag which we will use to build up the request to the broker      */
   /***************************************************************************/
   if (rc == OK)
   {
      mqCreateBag(MQCBO_COMMAND_BAG, &bag, &compCode, &reason);
      mqCheck("mqCreateBag", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Set command in bag                                                      */
   /***************************************************************************/
   if (rc == OK)
   {
      mqSetInteger(bag, MQIASY_COMMAND, 0, MQCMD_DEREGISTER_SUBSCRIBER,
                   &compCode, &reason);
      mqCheck("mqSetInteger", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Add Topic to bag                                                        */
   /***************************************************************************/
   if (rc == OK)
   {
      mqAddString(bag, MQCACF_TOPIC, MQBL_NULL_TERMINATED, Topic,
                  &compCode, &reason);
      mqCheck("mqAddString", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Add RegistrationOptions to bag                                          */
   /***************************************************************************/
   if (rc == OK)
   {
      regOptions = MQREGO_NONE;
      mqAddInteger(bag, MQIACF_REGISTRATION_OPTIONS, regOptions,
                   &compCode, &reason);
   } /* endif */

   /***************************************************************************/
   /* Add StreamName to bag                                                   */
   /***************************************************************************/
   if (rc == OK)
   {
      mqAddString(bag, MQCACF_STREAM_NAME, MQBL_NULL_TERMINATED, StreamName,
                  &compCode, &reason);
      mqCheck("mqAddString", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Add QMgrName to bag                                                     */
   /***************************************************************************/
   if (rc == OK)
   {
      mqAddString(bag, MQCA_Q_MGR_NAME, MQBL_NULL_TERMINATED, QMgrName,
                  &compCode, &reason);
      mqCheck("mqAddString", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Add QName to bag                                                        */
   /***************************************************************************/
   if (rc == OK)
   {
      mqAddString(bag, MQCA_Q_NAME, MQBL_NULL_TERMINATED, QName,
                  &compCode, &reason);
      mqCheck("mqAddString", compCode, reason);
   } /* endif */

   if (rc == OK)
       rc = SendBrokerCmd("Deregister Subscriber", &hCQObj, bag, DEFAULT_WAIT_INTERVAL);

   /***************************************************************************/
   /* If we try and deregister a subscriber and they are not registered it    */
   /* must be because the subscription expired.                               */
   /***************************************************************************/
   if( rc == MQRCCF_NOT_REGISTERED )
     rc = OK;

   /***************************************************************************/
   /* Delete the bag if we managed to create one                              */
   /***************************************************************************/
   if (bag != MQHB_UNUSABLE_HBAG)
   {
      mqDeleteBag(&bag, &compCode, &reason);
      mqCheck("mqDeleteBag", compCode, reason);
   } /* endif */

   return(rc);
}


/******************************************************************************/
/*                                                                            */
/* Function: ReqUpdate                                                        */
/*                                                                            */
/* Description: Request Update                                                */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Implements the MQCMD_REQUEST_UPDATE command message.                       */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* PMQCHAR Topic                  - topic                                     */
/* PMQCHAR StreamName             - stream name                               */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* None                                                                       */
/*                                                                            */
/******************************************************************************/
MQLONG ReqUpdate(PMQCHAR Topic, PMQCHAR StreamName)
{
   MQLONG rc=OK;                                 /* return code               */
   MQHBAG bag=MQHB_UNUSABLE_HBAG;                /* MQ bag                    */
   MQLONG reason;                                /* reason code               */
   MQLONG compCode;                              /* completion code           */
   MQLONG regOptions=0;                          /* registration options      */

   /***************************************************************************/
   /* Create bag which we will use to build up the request to the broker      */
   /***************************************************************************/
   if (rc == OK)
   {
      mqCreateBag(MQCBO_COMMAND_BAG, &bag, &compCode, &reason);
      mqCheck("mqCreateBag", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Set command in bag                                                      */
   /***************************************************************************/
   if (rc == OK)
   {
      mqSetInteger(bag, MQIASY_COMMAND, 0, MQCMD_REQUEST_UPDATE,
                   &compCode, &reason);
      mqCheck("mqSetInteger", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Add Topic to bag                                                        */
   /***************************************************************************/
   if (rc == OK)
   {
      mqAddString(bag, MQCACF_TOPIC, MQBL_NULL_TERMINATED, Topic,
                  &compCode, &reason);
      mqCheck("mqAddString", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Add RegistrationOptions to bag                                          */
   /***************************************************************************/
   if (rc == OK)
   {
      regOptions = MQREGO_NONE;
      mqAddInteger(bag, MQIACF_REGISTRATION_OPTIONS, regOptions,
                   &compCode, &reason);
   } /* endif */

   /***************************************************************************/
   /* Add StreamName to bag                                                   */
   /***************************************************************************/
   if (rc == OK)
   {
      mqAddString(bag, MQCACF_STREAM_NAME, MQBL_NULL_TERMINATED, StreamName,
                  &compCode, &reason);
      mqCheck("mqAddString", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Add QMgrName to bag                                                     */
   /***************************************************************************/
   if (rc == OK)
   {
      mqAddString(bag, MQCA_Q_MGR_NAME, MQBL_NULL_TERMINATED, QMgrName,
                  &compCode, &reason);
      mqCheck("mqAddString", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Add QName to bag                                                        */
   /***************************************************************************/
   if (rc == OK)
   {
      mqAddString(bag, MQCA_Q_NAME, MQBL_NULL_TERMINATED, QName,
                  &compCode, &reason);
      mqCheck("mqAddString", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Send the request update command to the broker and wait for a reply.     */
   /* (The reply will be sent after all the retained publications, which can  */
   /* take some time for a large request so a long wait may be required.)     */
   /***************************************************************************/
   if (rc == OK)
       rc = SendBrokerCmd("Request Update", &hCQObj, bag, EXTENDED_WAIT_INTERVAL);

   /***************************************************************************/
   /* Delete the bag if we managed to create one                              */
   /***************************************************************************/
   if (bag != MQHB_UNUSABLE_HBAG)
   {
      mqDeleteBag(&bag, &compCode, &reason);
      mqCheck("mqDeleteBag", compCode, reason);
   } /* endif */

   return(rc);
}


/******************************************************************************/
/*                                                                            */
/* Function: SendBrokerCmd                                                    */
/*                                                                            */
/* Description: Send broker command                                           */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Sends the broker the command message.                                      */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* PMQCHAR CmdStr                 - command string                            */
/* PMQHOBJ phObj;                 - ptr to handle to object                   */
/* MQHBAG  hBag                   - handle to bag                             */
/* MQLONG  WaitTime               - WaitInterval for mqGetBag                 */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                - returns from MQAI                         */
/*                                                                            */
/******************************************************************************/
MQLONG SendBrokerCmd(PMQCHAR CmdStr,
                     PMQHOBJ phObj,
                     MQHBAG  hBag,
                     MQLONG  WaitTime)
{
   MQLONG rc=OK;                                 /* return code               */
   MQLONG rc2=OK;                                /* return code 2             */
   MQLONG count;                                 /* used for loop counting    */
   MQLONG reason;                                /* reason code               */
   MQLONG compCode;                              /* completion code           */
   MQLONG compCode2;                             /* completion code 2         */
   MQCHAR buffer[BUFFER_SIZE]="";                /* buffer                    */
   MQMD putMD=md;                                /* put message descriptor    */
   MQMD getMD={MQMD_DEFAULT};                    /* get message descriptor    */
   MQPMO putMO={MQPMO_DEFAULT};                  /* put message options       */
   MQGMO getMO={MQGMO_DEFAULT};                  /* get message options       */
   MQHBAG replyBag=MQHB_UNUSABLE_HBAG;           /* reply bag                 */
   MQLONG itemValue=MQCFC_NOT_LAST;              /* item value                */
   MQLONG itemCount;                             /* item count                */
   MQLONG OutSel,OutType;                        /* selector outs             */
   MQCHAR strData[BUFFER_SIZE]="";               /* string data               */
   MQLONG strDataLen=sizeof(strData);            /* string data length        */
   MQLONG dataLen;                               /* data length received      */
   MQLONG TotalWait;                             /* Wait Interval.            */

   /***************************************************************************/
   /* Create reply bag                                                        */
   /***************************************************************************/
   if (rc == OK)
   {
      mqCreateBag(MQCBO_COMMAND_BAG, &replyBag, &compCode, &reason);
      mqCheck("mqCreateBag", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* Send the broker the command                                             */
   /***************************************************************************/
   if (rc == OK)
   {
      mqPutBag(hCon, *phObj, &putMD, &putMO, hBag, &compCode, &reason);
      mqCheck("mqPutBag", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* If the request was put successfully then loop round receiving replies   */
   /* until the last one is received or we get bored waiting.                 */
   /***************************************************************************/
   if (rc == OK)
   {
      getMO.Options = MQGMO_WAIT + MQGMO_CONVERT;

      while (rc == OK && itemValue != MQCFC_LAST)
      {
         memcpy(getMD.MsgId, MQMI_NONE, MQ_MSG_ID_LENGTH);
         memcpy(getMD.CorrelId, putMD.MsgId, MQ_MSG_ID_LENGTH);

         /*********************************************************************/
         /* Split a very long WaitInterval into reasonable segments and       */
         /* display status messages if there is a long delay.                 */
         /*********************************************************************/
         reason = MQRC_NO_MSG_AVAILABLE;
         TotalWait = WaitTime;
         for (count=0; reason == MQRC_NO_MSG_AVAILABLE && TotalWait > 0; count++)
         {
            switch (count)
            {
               case 0:
                  break;
               case 1:
                  fprintf(pLogFile, "Waiting for a response from a %s command.", CmdStr);
                  break;
               default:
                  fprintf(pLogFile, ".");
                  fflush(pLogFile);              /* force the output          */
            } /* endswitch */
            getMO.WaitInterval = min(TotalWait, WAIT_INTERVAL_SEGMENT);
            TotalWait -= WAIT_INTERVAL_SEGMENT;
            /******************************************************************/
            /* Get the reply message into the reply bag.                      */
            /******************************************************************/
            mqGetBag(hCon, hRQObj, &getMD, &getMO, replyBag, &compCode, &reason);
         } /* endfor */

         if (count > 1)
            fprintf(pLogFile, "\n");
         mqCheck("mqGetBag", compCode, reason);

         /*********************************************************************/
         /* Check that this is the reply message                              */
         /*********************************************************************/
         if (rc == OK)
         {
            if (getMD.MsgType != MQMT_REPLY)
            {
               fprintf(pLogFile, "Error expecting reply message, received message type: %d\n",
                                  getMD.MsgType);
               rc = NOT_OK;
            } /* endif */
         } /* endif */

         /*********************************************************************/
          /* Check the compCode and reason                                    */
         /*********************************************************************/
         if (rc == OK)
         {
            mqInquireInteger(replyBag, MQIASY_COMP_CODE, MQIND_NONE, &itemValue,
                             &compCode, &reason);
            mqCheck("mqInquireInteger", compCode, reason);

            compCode2 = itemValue;               /* save the compCode         */
         } /* endif */

         if (rc == OK)
         {
            mqInquireInteger(replyBag, MQIASY_REASON, MQIND_NONE, &itemValue,
                             &compCode, &reason);
            mqCheck("mqInquireInteger", compCode, reason);

            if (rc == OK)
            {
               /*-------------------------------------------------------------*/
               /* If we have already been round the loop and got an error,    */
               /* we will get a command failed error, however we want to      */
               /* continue until MQCFC_LAST is reached, to clear responses.   */
               /*-------------------------------------------------------------*/
               if (compCode2 != MQCC_OK && rc2 == OK)
               {
                  rc2 = itemValue;
                  /*----------------------------------------------------------*/
                  /* If the command is request update and there is not a      */
                  /* retained message, then that is not really an error. Also */
                  /* if we are deregistering a subscriber then it is possible */
                  /* the subscription has already expired, if so we can       */
                  /* accept the MQRCCF_NOT_REGISTERED error.                  */
                  /*----------------------------------------------------------*/
                  if (((strcmp(CmdStr, "Request Update") != 0) ||
                       (itemValue != MQRCCF_NO_RETAINED_MSG)) &&
                      ((strcmp(CmdStr, "Deregister Subscriber") != 0) ||
                       (itemValue != MQRCCF_NOT_REGISTERED)))
                  {
                     fprintf(pLogFile, "%s", CmdStr);
                     mqCheck("", compCode2, rc2);
                     rc = OK;                    /* reset rc from mqCheck     */
                  } /* endif */
               } /* endif */
            } /* endif */
         } /* endif */

         /*********************************************************************/
         /* Now print out any user selectors containing diagnostics           */
         /*********************************************************************/
         if (rc == OK)
         {
            mqCountItems(replyBag, MQSEL_ANY_USER_SELECTOR, &itemCount,
                         &compCode, &reason);
            mqCheck("mqCountItems", compCode, reason);

            if (rc == OK)
            {
               for (count=0; count < itemCount && rc == OK; count++)
               {
                  mqInquireItemInfo(replyBag, MQSEL_ANY_USER_SELECTOR, count,
                                    &OutSel, &OutType, &compCode, &reason);
                  mqCheck("mqInquireItemInfo", compCode, reason);

                  if (rc == OK)
                  {
                     if(OutType == MQIT_INTEGER)
                     {
                        mqInquireInteger(replyBag, MQSEL_ANY_USER_SELECTOR, count,
                                         &itemValue, &compCode, &reason);
                        mqCheck("mqInquireInteger", compCode, reason);

                        /*----------------------------------------------------*/
                        /* Only print message if it's a genuine               */
                        /*----------------------------------------------------*/
                        if (rc == OK)
                           fprintf(pLogFile, "Parameter: %d itemValue: %d\n",
                                   OutSel, itemValue);
                     }
                     else if(OutType == MQIT_STRING)
                     {
                        mqInquireString(replyBag, MQSEL_ANY_USER_SELECTOR, count,
                                        strDataLen, strData, &dataLen, NULL,
                                        &compCode, &reason);
                        mqCheck("mqInquireString", compCode, reason);

                        /*----------------------------------------------------*/
                        /* Only print message if it's a genuine               */
                        /*----------------------------------------------------*/
                        if (rc == OK)
                        {
                           strData[dataLen] = '\0';    /* terminate string    */
                           fprintf(pLogFile, "Parameter: %d itemValue: %s\n",
                                   OutSel, strData);
                        } /* endif */
                     } /* endif */
                  } /* endif */
               } /* endfor */
            } /* endif */
         } /* endif */

         /*********************************************************************/
         /* Exit the reply loop if the last reply message was received        */
         /*********************************************************************/
         if (rc == OK)
         {
            mqInquireInteger(replyBag, MQIASY_CONTROL, MQIND_NONE,
                             &itemValue, &compCode, &reason);
            mqCheck("mqInquireInteger", compCode, reason);
         } /* endif */
      } /* endwhile */
   } /* endif */

   /***************************************************************************/
   /* Delete the reply bag if we managed to create one                        */
   /***************************************************************************/
   if (replyBag != MQHB_UNUSABLE_HBAG)
   {
      mqDeleteBag(&replyBag, &compCode, &reason);
      mqCheck("mqDeleteBag", compCode, reason);
   } /* endif */

   /***************************************************************************/
   /* If all is OK return rc2, in case that was not OK                        */
   /***************************************************************************/
   if (rc == OK)
      rc = rc2;

   return(rc);
}


/******************************************************************************/
/*                                                                            */
/* Function: DumpStringData                                                   */
/*                                                                            */
/* Description: Dump the StringData of a messages received                    */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Dump the StringData of a messages received on queue QName                  */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* MQLONG Identifier              - the string identifier                     */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
MQLONG DumpStringData(MQLONG Identifier)
{
   MQLONG rc=OK;                                 /* return code               */
   MQLONG count;                                 /* used for loop counting    */
   MQLONG reason;                                /* reason code               */
   MQLONG compCode;                              /* completion code           */
   MQCHAR strData[BUFFER_SIZE]="";               /* string data               */
   MQLONG strDataLen;                            /* string data length        */
   MQLONG dataLen;                               /* data length received      */
   MQMD getMD={MQMD_DEFAULT};                    /* get message descriptor    */
   MQGMO getMO={MQGMO_DEFAULT};                  /* get message options       */
   MQHBAG pubBag=MQHB_UNUSABLE_HBAG;             /* bag for publishes         */
   MQLONG itemValue;                             /* item value                */
   MQLONG itemCount;                             /* item count                */

   /***************************************************************************/
   /* Create bag for the retained publishes to come into                      */
   /***************************************************************************/
   if (rc == OK)
   {
      mqCreateBag(MQCBO_COMMAND_BAG, &pubBag, &compCode, &reason);
      mqCheck("mqCreateBag", compCode, reason);
   } /* endif */

   if (rc == OK)
   {
      strDataLen = sizeof(strData) - 1;

      /************************************************************************/
      /* We have already received a reply from the request update command     */
      /* that requested these publication messages, as the reply message      */
      /* should be the last message sent all the publications should already  */
      /* be available to us, so no wait is required. It is possible that the  */
      /* queue manager is remote and of a different CCSID hence convert.      */
      /************************************************************************/
      getMO.Options = MQGMO_NO_WAIT + MQGMO_CONVERT;

      while (rc == OK && reason != MQRC_NO_MSG_AVAILABLE)
      {
         /*********************************************************************/
         /* In order to read the messages in sequence, MsgId & CorrelID must  */
         /* have the default value. MQGET sets them to the values for the     */
         /* message it returns, so re-initialise them before every call       */
         /*********************************************************************/
         memcpy(getMD.MsgId, MQMI_NONE, MQ_MSG_ID_LENGTH);
         memcpy(getMD.CorrelId, MQCI_NONE, MQ_CORREL_ID_LENGTH);

         /*********************************************************************/
         /* Clear the bag ready for a publish message to come into            */
         /*********************************************************************/
         mqClearBag(pubBag, &compCode, &reason);
         mqCheck("mqClearBag", compCode, reason);

         /*********************************************************************/
         /* Get the publish message into the bag                              */
         /*********************************************************************/
         mqGetBag(hCon, hRQObj, &getMD, &getMO, pubBag, &compCode, &reason);

         /*********************************************************************/
         /* If we got a message make sure it's OK then print StringData       */
         /*********************************************************************/
         if (reason != MQRC_NO_MSG_AVAILABLE)
         {
            mqCheck("mqGetBag", compCode, reason);

            /******************************************************************/
            /* Print publish message. If there is anything other than a       */
            /* publish on the queue, consider it an error, as I don't want    */
            /* anyone else using my queue.                                    */
            /******************************************************************/
            if (rc == OK)
            {
               mqInquireInteger(pubBag, MQIASY_COMMAND, MQIND_NONE,
                                &itemValue, &compCode, &reason);
               mqCheck("mqInquireInteger", compCode, reason);

               if (itemValue == MQCMD_PUBLISH)
               {
                  /*----------------------------------------------------------*/
                  /* Dump the string                                          */
                  /*----------------------------------------------------------*/
                  mqCountItems(pubBag, Identifier, &itemCount,
                               &compCode, &reason);
                  mqCheck("mqCountItems", compCode, reason);

                  if (rc == OK)
                  {
                     if (itemCount > 0)
                     {
                        for (count=0; count < itemCount && rc == OK; count++)
                        {
                           mqInquireString(pubBag, Identifier, count, strDataLen,
                                           strData, &dataLen, NULL, &compCode, &reason);
                          /*--------------------------------------------------*/
                          /* Don't care if the string is too big              */
                          /*--------------------------------------------------*/
                           if (reason != MQRC_STRING_TRUNCATED)
                              mqCheck("mqInquireString", compCode, reason);

                           if (rc == OK)
                           {
                              strData[dataLen] = '\0'; /* terminate string    */
                              fprintf(pLogFile, "   %s\n", strData);
                           } /* endif */
                        } /* endfor */
                     }
                     else
                     {
                        fprintf(pLogFile, "   None\n");
                     } /* endif */
                  } /* endif */
               }
               else
               {
                  fprintf(pLogFile, "Error non publish message found on queue: %s\n", QName);
               } /* endif */
            } /* endif */
         } /* endif */
      } /* endwhile */
   } /* endif */

   /***************************************************************************/
   /* Delete the publishes bag if we managed to create one                    */
   /***************************************************************************/
   if (pubBag != MQHB_UNUSABLE_HBAG)
   {
      mqDeleteBag(&pubBag, &compCode, &reason);
      mqCheck("mqDeleteBag", compCode, reason);
   } /* endif */

   return(rc);
}


/******************************************************************************/
/*                                                                            */
/* Function: DumpPubSubData                                                   */
/*                                                                            */
/* Description: Dump the Publisher or Subscriber data of a messages received  */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Dump the Publisher or Subscriber data of a messages received on queue      */
/* QName                                                                      */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* PMQCHAR SubStr                 - the string subscribed to without topic    */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
MQLONG DumpPubSubData(PMQCHAR SubStr)
{
   MQLONG rc=OK;                                 /* return code               */
   MQLONG count;                                 /* used for loop counting    */
   MQLONG reason;                                /* reason code               */
   MQLONG compCode;                              /* completion code           */
   MQCHAR strData[BUFFER_SIZE]="";               /* string data               */
   MQLONG strDataLen;                            /* string data length        */
   MQLONG dataLen;                               /* data length received      */
   MQLONG intData;                               /* integer data              */
   MQMD getMD={MQMD_DEFAULT};                    /* get message descriptor    */
   MQGMO getMO={MQGMO_DEFAULT};                  /* get message options       */
   MQHBAG pubBag=MQHB_UNUSABLE_HBAG;             /* bag for publishes         */
   MQLONG itemValue;                             /* item value                */
   MQLONG itemCount;                             /* item count                */
   MQLONG OutSel,OutType;                        /* selector outs             */

   /***************************************************************************/
   /* Create bag for the retained publishes to come into                      */
   /***************************************************************************/
   if (rc == OK)
   {
      mqCreateBag(MQCBO_COMMAND_BAG, &pubBag, &compCode, &reason);
      mqCheck("mqCreateBag", compCode, reason);
   } /* endif */

   if (rc == OK)
   {
      strDataLen = sizeof(strData) - 1;

      /************************************************************************/
      /* We have already received a reply from the request update command     */
      /* that requested these publication messages, as the reply message      */
      /* should be the last message sent all the publications should already  */
      /* be available to us, so no wait is required. It is possible that the  */
      /* queue manager is remote and of a different CCSID hence convert.      */
      /************************************************************************/
      getMO.Options = MQGMO_NO_WAIT + MQGMO_CONVERT;

      while (rc == OK && reason != MQRC_NO_MSG_AVAILABLE)
      {
         /*********************************************************************/
         /* In order to read the messages in sequence, MsgId & CorrelID must  */
         /* have the default value. MQGET sets them to the values for the     */
         /* message it returns, so re-initialise them before every call       */
         /*********************************************************************/
         memcpy(getMD.MsgId, MQMI_NONE, MQ_MSG_ID_LENGTH);
         memcpy(getMD.CorrelId, MQCI_NONE, MQ_CORREL_ID_LENGTH);

         /*********************************************************************/
         /* Clear the bag ready for a publish message to come into            */
         /*********************************************************************/
         mqClearBag(pubBag, &compCode, &reason);
         mqCheck("mqClearBag", compCode, reason);

         /*********************************************************************/
         /* Get the publish message into the bag                              */
         /*********************************************************************/
         mqGetBag(hCon, hRQObj, &getMD, &getMO, pubBag, &compCode, &reason);

         /*********************************************************************/
         /* If we got a message make sure it's OK then print StringData       */
         /*********************************************************************/
         if (reason != MQRC_NO_MSG_AVAILABLE)
         {
            mqCheck("mqGetBag", compCode, reason);

            /******************************************************************/
            /* Print publish message. If there is anything other than a       */
            /* publish on the queue, consider it an error, as I don't want    */
            /* anyone else using my queue.                                    */
            /******************************************************************/
            if (rc == OK)
            {
               mqInquireInteger(pubBag, MQIASY_COMMAND, MQIND_NONE,
                                &itemValue, &compCode, &reason);
               mqCheck("mqInquireInteger", compCode, reason);

               if (itemValue == MQCMD_PUBLISH)
               {
                  /************************************************************/
                  /* Pull everything out of the bag so we get things in the   */
                  /* correct sequence                                         */
                  /************************************************************/
                  mqCountItems(pubBag, MQSEL_ANY_SELECTOR, &itemCount,
                               &compCode, &reason);
                  mqCheck("mqCountItems", compCode, reason);

                  for (count=0; count < itemCount && rc == OK; count++)
                  {
                     mqInquireItemInfo(pubBag, MQSEL_ANY_SELECTOR, count,
                                       &OutSel, &OutType, &compCode, &reason);
                     mqCheck("mqInquireItemInfo", compCode, reason);

                     if (rc == OK)
                     {
                        /******************************************************/
                        /* Dump all the selectors we are interested in        */
                        /******************************************************/
                        switch (OutSel)
                        {
                           case MQCACF_TOPIC:

                              mqInquireString(pubBag, MQSEL_ANY_SELECTOR, count,
                                              strDataLen, strData, &dataLen, NULL,
                                              &compCode, &reason);
                              /*----------------------------------------------*/
                              /* Don't care if the string is too big          */
                              /*----------------------------------------------*/
                              if (reason != MQRC_STRING_TRUNCATED)
                                 mqCheck("mqInquireString", compCode, reason);

                              if (rc == OK)
                              {
                                 strData[dataLen] = '\0';  /* add terminator  */
                                 /*-------------------------------------------*/
                                 /* We have subscribed to SubStr\Topic so     */
                                 /* need to dump Topic without SubStr         */
                                 /*-------------------------------------------*/
                                 fprintf(pLogFile, "   Topic: %s\n", strData + strlen(SubStr));
                              } /* endif */

                              break;

                           case MQCACF_REG_Q_MGR_NAME:

                              rc = DumpPCFStringItemByIndex(pubBag, count,
                                                            "RegistrationQMgrName");
                              break;

                           case MQCACF_REG_Q_NAME:

                              rc = DumpPCFStringItemByIndex(pubBag, count,
                                                            "RegistrationQName");
                              break;

                           case MQCACF_REG_CORREL_ID:

                              rc = DumpPCFStringItemByIndex(pubBag, count,
                                                            "RegistrationCorrellId");
                              break;

                           case MQCACF_REG_USER_ID:

                              rc = DumpPCFStringItemByIndex(pubBag, count,
                                                            "RegistrationUserIdentifier");
                              break;

                           case MQIACF_REG_REG_OPTIONS:

                              mqInquireInteger(pubBag, MQSEL_ANY_SELECTOR, count,
                                               &intData, &compCode, &reason);
                              mqCheck("mqInquireInteger", compCode, reason);

                              if (rc == OK)
                              {
                                 fprintf(pLogFile, "      RegistrationOptions: %d :", intData);

                                 if (intData == MQREGO_NONE)
                                    fprintf(pLogFile, " MQREGO_NONE");
                                 if (intData & MQREGO_CORREL_ID_AS_IDENTITY)
                                    fprintf(pLogFile, " MQREGO_CORREL_ID_AS_IDENTITY");
                                 if (intData & MQREGO_ANONYMOUS)
                                    fprintf(pLogFile, " MQREGO_ANONYMOUS");
                                 if (intData & MQREGO_LOCAL)
                                    fprintf(pLogFile, " MQREGO_LOCAL");
                                 if (intData & MQREGO_DIRECT_REQUESTS)
                                    fprintf(pLogFile, " MQREGO_DIRECT_REQUESTS");
                                 if (intData & MQREGO_NEW_PUBLICATIONS_ONLY)
                                    fprintf(pLogFile, " MQREGO_NEW_PUBLICATIONS_ONLY");
                                 if (intData & MQREGO_PUBLISH_ON_REQUEST_ONLY)
                                    fprintf(pLogFile, " MQREGO_PUBLISH_ON_REQUEST_ONLY");
                                 if (intData & MQREGO_DEREGISTER_ALL)
                                    fprintf(pLogFile, " MQREGO_DEREGISTER_ALL");
                                 if (intData & MQREGO_INCLUDE_STREAM_NAME)
                                    fprintf(pLogFile, " MQREGO_INCLUDE_STREAM_NAME");
                                 if (intData & MQREGO_INFORM_IF_RETAINED)
                                    fprintf(pLogFile, " MQREGO_INFORM_IF_RETAINED");

                                 fprintf(pLogFile, "\n");
                              } /* endif */

                              break;

                           case MQCACF_REG_TIME:

                              rc = DumpPCFStringItemByIndex(pubBag, count,
                                                            "RegistrationTime");
                              break;

                           case MQIACF_BROKER_COUNT:

                              rc = DumpPCFIntegerItemByIndex(pubBag, count,
                                                             "BrokerCount");
                              break;

                           case MQIACF_ANONYMOUS_COUNT:

                              rc = DumpPCFIntegerItemByIndex(pubBag, count,
                                                             "AnonymousCount");
                              break;

                           case MQIACF_APPL_COUNT:

                              rc = DumpPCFIntegerItemByIndex(pubBag, count,
                                                             "ApplCount");
                              break;

                           default:              /* We are not interested in  */
                              break;             /* any other selectors       */
                        } /* endswitch */
                     } /* endif */
                  } /* endfor */
               }
               else
               {
                  fprintf(pLogFile, "Error non publish message found on queue: %s\n", QName);
               } /* endif */
            } /* endif */
         } /* endif */
      } /* endwhile */
   } /* endif */

   /***************************************************************************/
   /* Delete the publishes bag if we managed to create one                    */
   /***************************************************************************/
   if (pubBag != MQHB_UNUSABLE_HBAG)
   {
      mqDeleteBag(&pubBag, &compCode, &reason);
      mqCheck("mqDeleteBag", compCode, reason);
   } /* endif */

   return(rc);
}


/******************************************************************************/
/*                                                                            */
/* Function: DumpRetainedMessagesData                                         */
/*                                                                            */
/* Description: Dump the data of retained messages received                   */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Dump the data of retained messages received on queue QName                 */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* None                                                                       */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
MQLONG DumpRetainedMessagesData(void)
{
   MQLONG rc=OK;                                 /* return code               */
   MQLONG reason;                                /* reason code               */
   MQLONG compCode;                              /* completion code           */
   MQMD getMD={MQMD_DEFAULT};                    /* get message descriptor    */
   MQGMO getMO={MQGMO_DEFAULT};                  /* get message options       */
   PMQBYTE pMsgBuf;                              /* message buffer            */
   MQLONG msgBufLen=4096;                        /* message buffer length     */
   MQLONG msgLen;                                /* message length received   */

   pMsgBuf = malloc(msgBufLen);

   if (pMsgBuf == NULL)
   {
      fprintf(pLogFile, "Error malloc failed\n");
      rc = NOT_OK;
   } /* endif */

   if (rc == OK)
   {
      /************************************************************************/
      /* We have already received a reply from the request update command     */
      /* that requested these publication messages, as the reply message      */
      /* should be the last message sent all the publications should already  */
      /* be available to us, so no wait is required. It is possible that the  */
      /* queue manager is remote and of a different CCSID hence convert, this */
      /* may cause a problem if a user format of message is encountered.      */
      /* Depending upon whether or not the user format is known about via     */
      /* an exit, we may get MQRC_FORMAT_ERROR or MQRC_NOT_CONVERTED, in      */
      /* either case we will just dump the data.                              */
      /************************************************************************/
      getMO.Options = MQGMO_NO_WAIT + MQGMO_CONVERT;

      while (rc == OK && reason != MQRC_NO_MSG_AVAILABLE)
      {
         /*********************************************************************/
         /* In order to read the messages in sequence, MsgId & CorrelID must  */
         /* have the default value. MQGET sets them to the values for the     */
         /* message it returns, so re-initialise them before every call       */
         /*********************************************************************/
         memcpy(getMD.MsgId, MQMI_NONE, MQ_MSG_ID_LENGTH);
         memcpy(getMD.CorrelId, MQCI_NONE, MQ_CORREL_ID_LENGTH);

         /*********************************************************************/
         /* Get the publish message into the buffer.  Note DumpMessageData    */
         /* adds a string terminator, hence msgBufLen -1.                     */
         /*********************************************************************/
         MQGET(hCon, hRQObj, &getMD, &getMO, msgBufLen - 1, pMsgBuf, &msgLen,
               &compCode, &reason);

         /*-------------------------------------------------------------------*/
         /* If our buffer was not big enough, reallocate a bigger one         */
         /*-------------------------------------------------------------------*/
         if (reason == MQRC_TRUNCATED_MSG_FAILED)
         {
            msgBufLen = msgLen + 1;              /* 1 for string terminator   */
            pMsgBuf = realloc(pMsgBuf, msgBufLen);

            if (pMsgBuf != NULL)
            {
               MQGET(hCon, hRQObj, &getMD, &getMO, msgBufLen - 1, pMsgBuf, &msgLen,
                     &compCode, &reason);
            }
            else
            {
               fprintf(pLogFile, "Error realloc failed\n");
               rc = NOT_OK;
            } /* endif */
         } /* endif */

         /*********************************************************************/
         /* If we got a message make sure it's OK then print the data         */
         /*********************************************************************/
         if (reason != MQRC_NO_MSG_AVAILABLE)
         {
            if (reason == MQRC_NONE)
            {
               /***************************************************************/
               /* The message could be in PCF or RFH format                   */
               /***************************************************************/
               if (memcmp(getMD.Format, MQFMT_PCF, MQ_FORMAT_LENGTH) == 0)
               {
                  rc = DumpRetainedPCFMessageData(msgLen, pMsgBuf, getMD);
               }
               else if (memcmp(getMD.Format, MQFMT_RF_HEADER, MQ_FORMAT_LENGTH) == 0)
               {
                  rc = DumpRetainedRFHMessageData(msgLen, pMsgBuf, getMD);
               }
               else
               {
                  fprintf(pLogFile, "Format: %.*s message\n",
                                     MQ_FORMAT_LENGTH, getMD.Format);
                  DumpMessageData(msgLen, pMsgBuf);
               } /* endif */
            }
            else if (reason == MQRC_NOT_CONVERTED)
            {
               fprintf(pLogFile, "Warning - message not converted Format: %.*s message\n",
                                 MQ_FORMAT_LENGTH, getMD.Format);
               DumpMessageData(msgLen, pMsgBuf);
            }
            else if (reason == MQRC_FORMAT_ERROR)
            {
               fprintf(pLogFile, "Warning unknown Format: %.*s message\n",
                                  MQ_FORMAT_LENGTH, getMD.Format);
               DumpMessageData(msgLen, pMsgBuf);
            }
            else
            {
               mqCheck("MQGET", compCode, reason);
            } /* endif */
         } /* endif */
      } /* endwhile */
   } /* endif */

   /***************************************************************************/
   /* Free the memory                                                         */
   /***************************************************************************/
   if (pMsgBuf != NULL)
      free(pMsgBuf);

   return(rc);
}


/******************************************************************************/
/*                                                                            */
/* Function: DumpRetainedPCFMessageData                                       */
/*                                                                            */
/* Description: Dump the data of a retained PCF message                       */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Dump the data of a retained PCF message                                    */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* MQLONG msgLen                  - message length                            */
/* PMQBYTE pMsgBuf                - pointer to message buffer                 */
/* MQMD MD                        - message descriptor                        */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
MQLONG DumpRetainedPCFMessageData(MQLONG msgLen, PMQBYTE pMsgBuf, MQMD MD)
{
   MQLONG rc=OK;                                 /* return code               */
   MQLONG reason;                                /* reason code               */
   MQLONG compCode;                              /* completion code           */
   MQLONG intData;                               /* integer data              */
   MQHBAG pubBag=MQHB_UNUSABLE_HBAG;             /* bag for publishes         */
   MQLONG itemValue;                             /* item value                */
   MQLONG itemCount;                             /* item count                */

   /***************************************************************************/
   /* Create bag                                                              */
   /***************************************************************************/
   if (rc == OK)
   {
      mqCreateBag(MQCBO_COMMAND_BAG, &pubBag, &compCode, &reason);
      mqCheck("mqCreateBag", compCode, reason);
   } /* endif */

   if (rc == OK)
   {
     mqBufferToBag(MQHB_NONE, msgLen, pMsgBuf, pubBag, &compCode, &reason);
     mqCheck("mqBufferToBag", compCode, reason);
   } /* endif */

   if (rc == OK)
   {
      mqInquireInteger(pubBag, MQIASY_COMMAND, MQIND_NONE,
                       &itemValue, &compCode, &reason);
      mqCheck("mqInquireInteger", compCode, reason);
   } /* endif */

   if (rc == OK)
   {
      /************************************************************************/
      /* Print publish message. If there is anything other than a publish on  */
      /* the queue, consider it an error, as I don't want anyone else using   */
      /* my queue.                                                            */
      /************************************************************************/
      if (itemValue == MQCMD_PUBLISH)
      {
         mqCountItems(pubBag, MQIACF_PUBLICATION_OPTIONS, &itemCount,
                      &compCode, &reason);
         mqCheck("mqCountItems", compCode, reason);

         if (rc == OK)
         {
            if (itemCount == 1)
            {
               mqInquireInteger(pubBag, MQIACF_PUBLICATION_OPTIONS, 0,
                                &intData, &compCode, &reason);
               mqCheck("mqInquireInteger", compCode, reason);

               if (rc == OK)
               {
                  /*-------------------------------------------------------*/
                  /* All is well so dump the data we want                  */
                  /*-------------------------------------------------------*/
                  fprintf(pLogFile, "   PCF Message\n");
                  /*-------------------------------------------------------*/
                  /* Dump the expiry                                       */
                  /*-------------------------------------------------------*/
                  fprintf(pLogFile, "      Expiry: %d\n", MD.Expiry);

                  /*-------------------------------------------------------*/
                  /* Dump the Topic(s)                                     */
                  /*-------------------------------------------------------*/
                  if (rc == OK)
                     rc = DumpPCFStringItemBySelector(pubBag, MQCACF_TOPIC,
                                                      "Topic");

                  /*-------------------------------------------------------*/
                  /* Dump the QMgrName                                     */
                  /*-------------------------------------------------------*/
                  if (rc == OK)
                     rc = DumpPCFStringItemBySelector(pubBag, MQCA_Q_MGR_NAME,
                                                      "QMgrName");

                  /*-------------------------------------------------------*/
                  /* Dump the QName                                        */
                  /*-------------------------------------------------------*/
                  if (rc == OK)
                     rc = DumpPCFStringItemBySelector(pubBag, MQCA_Q_NAME,
                                                      "QName");

                  /*-------------------------------------------------------*/
                  /* Dump the SequenceNumber                               */
                  /*-------------------------------------------------------*/
                  if (rc == OK)
                     rc = DumpPCFIntegerItemBySelector(pubBag, MQIACF_SEQUENCE_NUMBER,
                                                       "SequenceNumber");

                  /*-------------------------------------------------------*/
                  /* Dump the PublishTimestamp                             */
                  /*-------------------------------------------------------*/
                  if (rc == OK)
                     rc = DumpPCFStringItemBySelector(pubBag, MQCACF_PUBLISH_TIMESTAMP,
                                                      "PublishTimestamp");

                  /*-------------------------------------------------------*/
                  /* Dump the IntegerData                                  */
                  /*-------------------------------------------------------*/
                  if (rc == OK)
                     rc = DumpPCFIntegerItemBySelector(pubBag, MQIACF_INTEGER_DATA,
                                                       "IntegerData");

                  /*-------------------------------------------------------*/
                  /* Dump the StringData                                   */
                  /*-------------------------------------------------------*/
                  if (rc == OK)
                     rc = DumpPCFStringItemBySelector(pubBag, MQCACF_STRING_DATA,
                                                      "StringData");
               } /* endif */
            }
            else
            {
               fprintf(pLogFile, "Error non retained message found on queue: %s\n", QName);
            } /* endif */
         } /* endif */
      }
      else
      {
         fprintf(pLogFile, "Error non publish message found on queue: %s\n", QName);
      } /* endif */
   } /* endif */

   /***************************************************************************/
   /* Delete the publishes bag if we managed to create one                    */
   /***************************************************************************/
   if (pubBag != MQHB_UNUSABLE_HBAG)
   {
      mqDeleteBag(&pubBag, &compCode, &reason);
      mqCheck("mqDeleteBag", compCode, reason);
   } /* endif */

   return(rc);
}


/******************************************************************************/
/*                                                                            */
/* Function: DumpRetainedRFHMessageData                                       */
/*                                                                            */
/* Description: Dump the data of a retained RFH message                       */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Dump the data of a retained RFH message                                    */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* MQLONG msgLen                  - message length                            */
/* PMQBYTE pMsgBuf                - pointer to message buffer                 */
/* MQMD MD                        - message descriptor                        */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
MQLONG DumpRetainedRFHMessageData(MQLONG msgLen, PMQBYTE pMsgBuf, MQMD MD)
{
   MQLONG rc=OK;                                 /* return code               */
   PMQCHAR pStr;                                 /* ptr to str                */
   MQLONG strLen;                                /* string length             */
   PMQCHAR pTagValue;                            /* ptr to tag value          */
   ULONG tagValueLen;                            /* tag value length          */
   BOOL tagFound;                                /* tag found boolean         */
   PMQRFH pRFH;                                  /* ptr to RFH                */
   PMQCHAR pRFHS;                                /* ptr to RFH name value str */
   MQLONG RFHSLen;                               /* RFH name value str length */
   PMQBYTE pRFHData;                             /* ptr to RFH data           */

   pRFH = (PMQRFH) pMsgBuf;
   pRFHS = (PMQCHAR) pMsgBuf + MQRFH_STRUC_LENGTH_FIXED;
   pRFHData = pMsgBuf + pRFH->StrucLength;
   RFHSLen = pRFH->StrucLength - MQRFH_STRUC_LENGTH_FIXED;

   /***************************************************************************/
   /* Print publish message. If there is anything other than a publish on the */
   /* queue, consider it an error, as I don't want anyone else using my queue */
   /***************************************************************************/
   if (rc == OK)
   {
      pStr = pRFHS;                              /* initialise ptr to str     */
      strLen = RFHSLen;                          /* initialise str length     */

      rc = GetRFHTagValue(&pStr, &strLen, MQPS_COMMAND, &pTagValue, &tagValueLen);

      if (rc == OK)
      {
         if (tagValueLen == 0)
         {
            fprintf(pLogFile, "Error no command in message found on queue: %s\n", QName);
            rc = NOT_OK;
         }
         else
         {
            if (tagValueLen != strlen(MQPS_PUBLISH) ||
                memcmp(pTagValue, MQPS_PUBLISH, strlen(MQPS_PUBLISH)) != 0)
            {
               fprintf(pLogFile, "Error non publish message found on queue: %s command:%.*s\n",
                                 QName, tagValueLen, pTagValue);
               rc = NOT_OK;
            } /* endif */
         } /* endif */
      } /* endif */
   } /* endif */

   /***************************************************************************/
   /* Dump the RFH message                                                    */
   /***************************************************************************/
   if (rc == OK)
   {
      fprintf(pLogFile, "   RFH Message\n");
      fprintf(pLogFile, "      Expiry: %d\n", MD.Expiry);

      /*----------------------------------------------------------------------*/
      /* Dump the Topic(s)                                                    */
      /*----------------------------------------------------------------------*/
      pStr = pRFHS;                              /* initialise ptr to str     */
      strLen = RFHSLen;                          /* initialise str length     */
      tagFound = FALSE;                          /* initialise found boolean  */

      do
      {
         rc = GetRFHTagValue(&pStr, &strLen, MQPS_TOPIC, &pTagValue, &tagValueLen);

         if (rc == OK && tagValueLen != 0)
         {
            tagFound = TRUE;
            fprintf(pLogFile, "      Topic: %.*s\n", tagValueLen, pTagValue);
         } /* endif */

         if (strLen == 0)                        /* end of name value str ?   */
            tagValueLen = 0;                     /* end loop                  */

      } while (rc == OK && tagValueLen != 0); /* enddo */

      /*----------------------------------------------------------------------*/
      /* Topic is mandatory in a publish message                              */
      /*----------------------------------------------------------------------*/
      if (!tagFound)
         fprintf(pLogFile, "Error Topic not found in RFH message\n");

      /*----------------------------------------------------------------------*/
      /* Dump the QMgrName                                                    */
      /*----------------------------------------------------------------------*/
      pStr = pRFHS;                              /* initialise ptr to str     */
      strLen = RFHSLen;                          /* initialise str length     */

      rc = GetRFHTagValue(&pStr, &strLen, MQPS_Q_MGR_NAME, &pTagValue, &tagValueLen);

      if (tagValueLen != 0)
         fprintf(pLogFile, "      QMgrName: %.*s\n", tagValueLen, pTagValue);
      else
         fprintf(pLogFile, "      QMgrName: None\n");

      /*----------------------------------------------------------------------*/
      /* Dump the QName                                                       */
      /*----------------------------------------------------------------------*/
      pStr = pRFHS;                              /* initialise ptr to str     */
      strLen = RFHSLen;                          /* initialise str length     */

      rc = GetRFHTagValue(&pStr, &strLen, MQPS_Q_NAME, &pTagValue, &tagValueLen);

      if (tagValueLen != 0)
         fprintf(pLogFile, "      QName: %.*s\n", tagValueLen, pTagValue);
      else
         fprintf(pLogFile, "      QName: None\n");

      /*----------------------------------------------------------------------*/
      /* Dump the SequenceNumber                                              */
      /*----------------------------------------------------------------------*/
      pStr = pRFHS;                              /* initialise ptr to str     */
      strLen = RFHSLen;                          /* initialise str length     */

      rc = GetRFHTagValue(&pStr, &strLen, MQPS_SEQUENCE_NUMBER, &pTagValue, &tagValueLen);

      if (tagValueLen != 0)
         fprintf(pLogFile, "      SequenceNumber: %.*s\n", tagValueLen, pTagValue);
      else
         fprintf(pLogFile, "      SequenceNumber: None\n");

      /*----------------------------------------------------------------------*/
      /* Dump the PublishTimestamp                                            */
      /*----------------------------------------------------------------------*/
      pStr = pRFHS;                              /* initialise ptr to str     */
      strLen = RFHSLen;                          /* initialise str length     */

      rc = GetRFHTagValue(&pStr, &strLen, MQPS_PUBLISH_TIMESTAMP, &pTagValue, &tagValueLen);

      if (tagValueLen != 0)
         fprintf(pLogFile, "      PublishTimestamp: %.*s\n", tagValueLen, pTagValue);
      else
         fprintf(pLogFile, "      PublishTimestamp: None\n");

   } /* endif */

   /***************************************************************************/
   /* Dump the message data                                                   */
   /***************************************************************************/
   if (rc == OK)
      DumpMessageData(msgLen - pRFH->StrucLength, pRFHData);

   return(rc);
}

/******************************************************************************/
/*                                                                            */
/* Function: BuildStreamList                                                  */
/*                                                                            */
/* Description: Builds the stream supported list                              */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Builds the stream supported list and dumps the stream names                */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* None                                                                       */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
MQLONG BuildStreamList(void)
{
   MQLONG rc=OK;                                 /* return code               */
   MQLONG count;                                 /* used for loop counting    */
   MQLONG reason;                                /* reason code               */
   MQLONG compCode;                              /* completion code           */
   MQCHAR strData[BUFFER_SIZE]="";               /* string data               */
   MQLONG strDataLen;                            /* string data length        */
   MQLONG dataLen;                               /* data length received      */
   MQMD getMD={MQMD_DEFAULT};                    /* get message descriptor    */
   MQGMO getMO={MQGMO_DEFAULT};                  /* get message options       */
   MQHBAG pubBag=MQHB_UNUSABLE_HBAG;             /* bag for publishes         */
   MQLONG itemValue;                             /* item value                */
   MQLONG itemCount;                             /* item count                */
   PStreamListEntry pNewSLE;                     /* -> new stream list entry  */

   /***************************************************************************/
   /* Create bag for the publishes to come into                               */
   /***************************************************************************/
   if (rc == OK)
   {
      mqCreateBag(MQCBO_COMMAND_BAG, &pubBag, &compCode, &reason);
      mqCheck("mqCreateBag", compCode, reason);
   } /* endif */

   if (rc == OK)
   {
      strDataLen = sizeof(strData) - 1;

      /************************************************************************/
      /* We have already received a reply from the request update command     */
      /* that requested these publication messages, as the reply message      */
      /* should be the last message sent all the publications should already  */
      /* be available to us, so no wait is required.                          */
      /************************************************************************/
      getMO.Options = MQGMO_NO_WAIT + MQGMO_CONVERT;

      /************************************************************************/
      /* We will only get one message for all the streams supported, so no    */
      /* need to loop round getting messages.                                 */
      /************************************************************************/

      /************************************************************************/
      /* In order to read the messages in sequence, MsgId & CorrelID must     */
      /* have the default value. MQGET sets them to the values for the        */
      /* message it returns, so re-initialise them before every call          */
      /************************************************************************/
      memcpy(getMD.MsgId, MQMI_NONE, MQ_MSG_ID_LENGTH);
      memcpy(getMD.CorrelId, MQCI_NONE, MQ_CORREL_ID_LENGTH);

      /************************************************************************/
      /* Clear the bag ready for a publish message to come into               */
      /************************************************************************/
      mqClearBag(pubBag, &compCode, &reason);
      mqCheck("mqClearBag", compCode, reason);

      /************************************************************************/
      /* Get the publish message into the bag                                 */
      /************************************************************************/
      mqGetBag(hCon, hRQObj, &getMD, &getMO, pubBag, &compCode, &reason);

      /************************************************************************/
      /* If we got a message make sure it's OK then print StringData          */
      /************************************************************************/
      if (reason != MQRC_NO_MSG_AVAILABLE)
      {
         mqCheck("mqGetBag", compCode, reason);

         /*********************************************************************/
         /* Print publish message. If there is anything other than a publish  */
         /* on the queue, consider it an error, as I don't want anyone else   */
         /* using my queue.                                                   */
         /*********************************************************************/
         if (rc == OK)
         {
            mqInquireInteger(pubBag, MQIASY_COMMAND, MQIND_NONE,
                             &itemValue, &compCode, &reason);
            mqCheck("mqInquireInteger", compCode, reason);

            if (itemValue == MQCMD_PUBLISH)
            {
               /*-------------------------------------------------------------*/
               /* Dump the string                                             */
               /*-------------------------------------------------------------*/
               mqCountItems(pubBag, MQCACF_SUPPORTED_STREAM_NAME, &itemCount,
                            &compCode, &reason);
               mqCheck("mqCountItems", compCode, reason);

               if (rc == OK)
               {
                  if (itemCount > 0)
                  {
                     for (count=0; count < itemCount && rc == OK; count++)
                     {
                        mqInquireString(pubBag, MQCACF_SUPPORTED_STREAM_NAME,
                                        count, strDataLen, strData, &dataLen,
                                        NULL, &compCode, &reason);
                        mqCheck("mqInquireString", compCode, reason);

                        if (rc == OK)
                        {
                           strData[dataLen] = '\0'; /* terminate string    */
                           fprintf(pLogFile, "%s\n", strData);

                           pNewSLE = malloc(sizeof(StreamListEntry));

                           if (pNewSLE != NULL)
                           {
                              strcpy(pNewSLE->StreamName, strData);
                              pNewSLE->pNextSLE = pStreamList;
                              pStreamList = pNewSLE;
                           }
                           else
                           {
                              fprintf(pLogFile, "Error malloc failed !\n");
                              rc = NOT_OK;
                           } /* endif */
                        } /* endif */
                     } /* endfor */
                  }
                  else
                  {
                     fprintf(pLogFile, "None\n");
                  } /* endif */
               } /* endif */
            }
            else
            {
               fprintf(pLogFile, "Error non publish message found on queue: %s\n", QName);
            } /* endif */
         } /* endif */
      } /* endif */
   } /* endif */

   /***************************************************************************/
   /* Delete the publishes bag if we managed to create one                    */
   /***************************************************************************/
   if (pubBag != MQHB_UNUSABLE_HBAG)
   {
      mqDeleteBag(&pubBag, &compCode, &reason);
      mqCheck("mqDeleteBag", compCode, reason);
   } /* endif */

   return(rc);
}


/******************************************************************************/
/*                                                                            */
/* Function: FreeStreamList                                                   */
/*                                                                            */
/* Description: Frees the stream supported list                               */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Frees the stream supported list                                            */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* None                                                                       */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
MQLONG FreeStreamList(void)
{
   MQLONG rc=OK;                                 /* return code               */
   PStreamListEntry pSLE;                        /* -> stream list entry      */
   PStreamListEntry pNextSLE;                    /* -> next stream list entry */

   for (pSLE=pStreamList; pSLE != NULL; pSLE=pNextSLE)
   {
      pNextSLE = pSLE->pNextSLE;                 /* get next SLE pointer      */
      free(pSLE);
   } /* endfor */

   pStreamList = NULL;                           /* set global pointer        */

   return(rc);
}


/******************************************************************************/
/*                                                                            */
/* Function: DumpPCFIntegerItemBySelector                                     */
/*                                                                            */
/* Description: Dumps a PCF integer item                                      */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Inquires from a bag an integer item selected by a selector, then dumps the */
/* information                                                                */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* MQHBAG Bag                     - handle to bag                             */
/* MQLONG Selector                - selector                                  */
/* char *Title                    - title to accompany information            */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
MQLONG DumpPCFIntegerItemBySelector(MQHBAG Bag, MQLONG Selector, char *Title)
{
   MQLONG rc=OK;                                 /* return code               */
   MQLONG count;                                 /* used for loop counting    */
   MQLONG reason;                                /* reason code               */
   MQLONG compCode;                              /* completion code           */
   MQLONG itemCount;                             /* item count                */
   MQLONG intData;                               /* integer data              */

   mqCountItems(Bag, Selector, &itemCount, &compCode, &reason);
   mqCheck("mqCountItems", compCode, reason);

   if (rc == OK)
   {
      if (itemCount > 0)
      {
         for (count=0; count < itemCount && rc == OK; count++)
         {
            mqInquireInteger(Bag, Selector, count, &intData, &compCode, &reason);
            mqCheck("mqInquireInteger", compCode, reason);
            if (rc == OK)
               fprintf(pLogFile, "      %s: %d\n", Title, intData);
         } /* endfor */
      }
      else
      {
         fprintf(pLogFile, "      %s: None\n", Title);
      } /* endif */
   } /* endif */

   return(rc);
}


/******************************************************************************/
/*                                                                            */
/* Function: DumpPCFStringItemBySelector                                      */
/*                                                                            */
/* Description: Dumps a PCF string item                                       */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Inquires from a bag a string item selected by a selector, then dumps the   */
/* information                                                                */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* MQHBAG Bag                     - handle to bag                             */
/* MQLONG Selector                - selector                                  */
/* char *Title                    - title to accompany information            */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
MQLONG DumpPCFStringItemBySelector(MQHBAG Bag, MQLONG Selector, char *Title)
{
   MQLONG rc=OK;                                 /* return code               */
   MQLONG count;                                 /* used for loop counting    */
   MQLONG reason;                                /* reason code               */
   MQLONG compCode;                              /* completion code           */
   MQLONG itemCount;                             /* item count                */
   MQCHAR strData[BUFFER_SIZE]="";               /* string data               */
   MQLONG strDataLen;                            /* string data length        */
   MQLONG dataLen;                               /* data length received      */

   strDataLen = sizeof(strData) - 1;             /* allow 1 for string term   */

   mqCountItems(Bag, Selector, &itemCount, &compCode, &reason);
   mqCheck("mqCountItems", compCode, reason);

   if (rc == OK)
   {
      if (itemCount > 0)
      {
         for (count=0; count < itemCount && rc == OK; count++)
         {
            mqInquireString(Bag, Selector, count, strDataLen, strData, &dataLen,
                            NULL, &compCode, &reason);
            /*----------------------------------------------------------------*/
            /* Don't care if the string is too big                            */
            /*----------------------------------------------------------------*/
            if (reason != MQRC_STRING_TRUNCATED)
               mqCheck("mqInquireString", compCode, reason);

            if (rc == OK)
            {
               strData[dataLen] = '\0';          /* add terminator            */
               fprintf(pLogFile, "      %s: %s\n", Title, strData);
            } /* endif */
         } /* endfor */
      }
      else
      {
         fprintf(pLogFile, "      %s: None\n", Title);
      } /* endif */
   } /* endif */

   return(rc);
}


/******************************************************************************/
/*                                                                            */
/* Function: DumpPCFIntegerItemByIndex                                        */
/*                                                                            */
/* Description: Dumps a PCF integer item                                      */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Inquires from a bag an integer item selected by an index, then dumps the   */
/* information                                                                */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* MQHBAG Bag                     - handle to bag                             */
/* MQLONG ItemIndex               - item index                                */
/* char *Title                    - title to accompany information            */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
MQLONG DumpPCFIntegerItemByIndex(MQHBAG Bag, MQLONG ItemIndex, char *Title)
{
   MQLONG rc=OK;                                 /* return code               */
   MQLONG reason;                                /* reason code               */
   MQLONG compCode;                              /* completion code           */
   MQLONG intData;                               /* integer data              */

   mqInquireInteger(Bag, MQSEL_ANY_SELECTOR, ItemIndex, &intData,
                    &compCode, &reason);
   mqCheck("mqInquireInteger", compCode, reason);

   if (rc == OK)
      fprintf(pLogFile, "      %s: %d\n", Title, intData);

   return(rc);
}


/******************************************************************************/
/*                                                                            */
/* Function: DumpPCFStringItemByIndex                                         */
/*                                                                            */
/* Description: Dumps a PCF string item                                       */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Inquires from a bag a string item selected by an index, then dumps the     */
/* information                                                                */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* MQHBAG Bag                     - handle to bag                             */
/* MQLONG ItemIndex               - item index                                */
/* char *Title                    - title to accompany information            */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* OK                             - all is well                               */
/*                                                                            */
/******************************************************************************/
MQLONG DumpPCFStringItemByIndex(MQHBAG Bag, MQLONG ItemIndex, char *Title)
{
   MQLONG rc=OK;                                 /* return code               */
   MQLONG reason;                                /* reason code               */
   MQLONG compCode;                              /* completion code           */
   MQCHAR strData[BUFFER_SIZE]="";               /* string data               */
   MQLONG strDataLen;                            /* string data length        */
   MQLONG dataLen;                               /* data length received      */

   strDataLen = sizeof(strData) - 1;             /* allow 1 for string term   */

   mqInquireString(Bag, MQSEL_ANY_SELECTOR, ItemIndex, strDataLen, strData,
                   &dataLen, NULL, &compCode, &reason);
   /*-------------------------------------------------------------------------*/
   /* Don't care if the string is too big                                     */
   /*-------------------------------------------------------------------------*/
   if (reason != MQRC_STRING_TRUNCATED)
      mqCheck("mqInquireString", compCode, reason);

   if (rc == OK)
   {
      strData[dataLen] = '\0';                   /* add terminator            */
      fprintf(pLogFile, "      %s: %s\n", Title, strData);
   } /* endif */

   return(rc);
}


#if MQAT_DEFAULT == MQAT_WINDOWS_NT || MQAT_DEFAULT == MQAT_OS2
/******************************************************************************/
/*                                                                            */
/* Function: getopt                                                           */
/*                                                                            */
/* Description: Get options                                                   */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Retrieves options and option-arguments from a list of parameters.          */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* int argc                       - number of arguments supplied              */
/* char *argv[]                   - program arguments                         */
/* char *opts                     - valid options                             */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* int c                          - option character                          */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Note: This code is taken from AMQSBLST.C, see for more details of this     */
/*       codes origin.                                                        */
/*                                                                            */
/******************************************************************************/
int getopt(int argc, char *argv[], char *opts)
{
   static int sp = 1;
   register int c;
   register char *cp;

   if(sp == 1)
   {
      if(optind >= argc ||  argv[optind][1] == '\0' ||
         (argv[optind][0] != '-' && argv[optind][0] != '/'))
      {
         return(EOF);
      }
      else if(strcmp(argv[optind], "--") == 0 ||
              strcmp(argv[optind], "//") == 0)
      {
         optind++;
         return(EOF);
      }
   }

   optopt = c = argv[optind][sp];

   if(c == ':' || (cp=strchr(opts, c)) == NULL)
   {
      if(argv[optind][++sp] == '\0')
      {
         optind++;
         sp = 1;
      }
      return('?');
   }

   if(*++cp == ':')
   {
      if(argv[optind][sp+1] != '\0')
      {
         optarg = &argv[optind++][sp+1];
      }
      else if(++optind >= argc)
      {
         sp = 1;
         return('?');
      }
      else
      {
         optarg = argv[optind++];
      }

      sp = 1;
   }
   else
   {
      if(argv[optind][++sp] == '\0')
      {
         sp = 1;
         optind++;
      }
      optarg = NULL;
   }
   return(c);
}
#endif


/******************************************************************************/
/*                                                                            */
/* Function: DumpMessageData                                                  */
/*                                                                            */
/* Description: Dump message data                                             */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/* ---------                                                                  */
/* Dumps messsage data, format can be of any type.                            */
/*                                                                            */
/* Input Parameters:                                                          */
/* -----------------                                                          */
/* MQLONG DataLength              - message data length                       */
/* PMQBYTE Buffer                 - message data buffer                       */
/*                                                                            */
/* In/Out Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Output Parameters:                                                         */
/* ------------------                                                         */
/* None                                                                       */
/*                                                                            */
/* Returns:                                                                   */
/* --------                                                                   */
/* None                                                                       */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Note: This code is taken from AMQSBCG0.C                                   */
/*                                                                            */
/******************************************************************************/
void DumpMessageData(MQLONG DataLength, PMQBYTE Buffer)
{
  int i = 0;       /* loop counter                                    */

  /* variables for message formatting *****/
  int  ch;
  int  overrun;  /* used on MBCS characters */
  int  mbcsmax;  /* used for MBCS characters */
  int  char_len;  /* used for MBCS characters */
  char line_text[CHARS_PER_LINE + 4]; /* allows for up to 3 MBCS bytes overrun */
  int  chars_this_line = 0;
  int  lines_printed   = 0;

  /*                                       */
  /* determine locale for MBCS handling    */
  /*                                       */
  setlocale(LC_ALL,"");  /* for mbcs charactersets */
  mbcsmax = MB_CUR_MAX;  /* for mbcs charactersets */

  fprintf(pLogFile, "\n      ****   Message   ****");

  /* only dump messsages up to MAX_MSG_DUMPED_SIZE */
  if (DataLength > MAX_MSG_DUMPED_SIZE)
  {
    fprintf(pLogFile, "   length - %d bytes - %d bytes dumped\n ",
                      DataLength, MAX_MSG_DUMPED_SIZE);
    DataLength = MAX_MSG_DUMPED_SIZE;
  }
  else
  {
     fprintf(pLogFile, "   length - %d bytes\n ", DataLength);
  } /* endif */

  Buffer[DataLength] = '\0';

  ch = 0;
  overrun = 0;
  do
  {
    chars_this_line = 0;
    fprintf(pLogFile, "\n      %08X: ",ch);
    for (;overrun>0; overrun--) /* for MBCS overruns */
    {
      fprintf(pLogFile, "  ");            /* dummy space for characters  */
      line_text[chars_this_line] = ' ';
                           /* included in previous line */
      chars_this_line++;
      if (overrun % 2)
        fprintf(pLogFile, " ");
    }
    while ( (chars_this_line < CHARS_PER_LINE) &&
            (ch < DataLength) )
    {
      char_len = mblen((char *)&Buffer[ch],mbcsmax);
      if (char_len < 1)   /* badly formed mbcs character */
        char_len = 1;     /* or NULL treated as sbcs     */
      if (char_len > 1 )
      { /* mbcs case, assumes mbcs are all printable */
        for (;char_len >0;char_len--)
        {
          if ((chars_this_line % 2 == 0) &&
              (chars_this_line < CHARS_PER_LINE))
            fprintf(pLogFile, " ");
          fprintf(pLogFile, "%02X",Buffer[ch] );
          line_text[chars_this_line] = Buffer[ch];
          chars_this_line++;
          ch++;
        }
      }
      else
      {  /* sbcs case */
        if (chars_this_line % 2 == 0)
          fprintf(pLogFile, " ");
        fprintf(pLogFile, "%02X",Buffer[ch] );
        line_text[chars_this_line] =
            isprint(Buffer[ch]) ? Buffer[ch] : '.';
        chars_this_line++;
        ch++;
      }
    }

    /* has an mbcs character overun the usual end? */
    if (chars_this_line > CHARS_PER_LINE)
       overrun = chars_this_line - CHARS_PER_LINE;

    /* pad with blanks to format the last line correctly */
    if (chars_this_line < CHARS_PER_LINE)
    {
      for ( ;chars_this_line < CHARS_PER_LINE;
           chars_this_line++)
      {
        if (chars_this_line % 2 == 0) fprintf(pLogFile, " ");
        fprintf(pLogFile, "  ");
        line_text[chars_this_line] = ' ';
      }
    }

    /* leave extra space between colums if MBCS characters possible */
    for (i=0;i < ((mbcsmax - overrun - 1) *2);i++)
    {
      fprintf(pLogFile, " "); /* prints space between hex represention and character */
    }

    line_text[chars_this_line] = '\0';
    fprintf(pLogFile, " '%s'",line_text);
    lines_printed += 1;
    if (lines_printed >= 60)
    {
      lines_printed = 0;
      fprintf(pLogFile, "\n ");
    }
  }
  while (ch < DataLength);

  fprintf(pLogFile, "\n");
}

/*********************************************************************/
/*                                                                   */
/* Function Name : GetRFHTagValue                                    */
/*                                                                   */
/* Description   : Get a tag value in a NameValueString              */
/*                                                                   */
/* Flow          :                                                   */
/*                                                                   */
/*  While we have not reached the end of the NameValueString or      */
/*  until we find the tag:                                           */
/*   GetNextToken from the string (name)                             */
/*   GetNextToken from the string (value)                            */
/*   If the name token is tag name process the value                 */
/*  end while                                                        */
/*                                                                   */
/*********************************************************************/
/*                                                                   */
/* Note: This code is taken from AMQSRESA.C, it has been modified    */
/*       so that it looks for any tag rather than a topic.           */
/*                                                                   */
/*********************************************************************/
MQLONG GetRFHTagValue(PMQCHAR *ppNameValueString,
                      MQLONG  *pStringLength,
                      PMQCHAR  pTagName,
                      PMQCHAR *ppTagValue,
                      ULONG   *pTagValueLength)
{
  MQLONG  rc = OK;
  BOOL    bTagFound = FALSE;
  PMQBYTE pTag;
  ULONG   tagLength;
  PMQBYTE pValue;
  ULONG   valueLength;

  *pTagValueLength = 0;

  /*******************************************************************/
  /* Search along the NameValueString until we find the tag          */
  /* name token, or we get to the end of the string.                 */
  /*******************************************************************/
  while( (rc == OK)
       &&(*pStringLength > 0)
       &&(bTagFound == FALSE) )
  {
    /*****************************************************************/
    /* Locate the next blank delimited token in the string. The      */
    /* tokens always occur in pairs so this will always be a name    */
    /* token and the following token its value.                      */
    /*****************************************************************/
    rc = GetNextToken( ppNameValueString
                     , pStringLength
                     , &pTag
                     , &tagLength );

    /*****************************************************************/
    /* If a token was found continue, otherwise we have reached the  */
    /* end of the string.                                            */
    /*****************************************************************/
    if( (rc == OK)
      &&(tagLength > 0) )
    {
      /***************************************************************/
      /* Locate the previous token's value.                          */
      /***************************************************************/
      rc = GetNextToken( ppNameValueString
                       , pStringLength
                       , &pValue
                       , &valueLength );
      if( rc == OK )
      {
        /*************************************************************/
        /* If a token was found continue, otherwise we have a name   */
        /* token with no value, this is an error.                    */
        /*************************************************************/
        if( valueLength == 0 )
        {
          fprintf(pLogFile, "Odd number of tokens found in NameValueString\n");
          rc = NOT_OK;
        }
        /*************************************************************/
        /* If the name token is pTagName we have found what we       */
        /* were looking for and we must parse the tag value.         */
        /*************************************************************/
        else if( (tagLength == strlen(pTagName))
               &&(memcmp(pTag, pTagName, strlen(pTagName)) == 0) )
        {
          bTagFound = TRUE;
          /***********************************************************/
          /* If the first character of the tag value returned by     */
          /* the tokenizer is a '"' then the tag is contained in     */
          /* quotes and these can be stripped off.                   */
          /* Note: It is known that no publications we are expecting */
          /*       will contain embedded quotes so there is no       */
          /*       requirement here to remove them. In a general     */
          /*       system replacement of '""' with '"' may be        */
          /*       required.                                         */
          /***********************************************************/
          if( *pValue == '"' )
          {
            pValue++;
            valueLength -= 2;
          }

          *ppTagValue = (PMQCHAR) pValue;
          *pTagValueLength = valueLength;
        }
      }
    }
  }

  return(rc);
}
/*********************************************************************/
/* end of ExtractRFHTagValue                                         */
/*********************************************************************/


/*********************************************************************/
/*                                                                   */
/* Function Name : GetNextToken                                      */
/*                                                                   */
/* Description   : Parse the supplied string for the start and end   */
/*                 of a valid token.                                 */
/*                                                                   */
/* Note          : This function is a generic function that can be   */
/*                 used in any application that needs to parse the   */
/*                 NameValueString of an MQRFH message.              */
/*                                                                   */
/* Flow          :                                                   */
/*                                                                   */
/*  While we have not reached the end of the string or a token:      */
/*   Process the current character:                                  */
/*    '\0'  - End of string                                          */
/*    ' '   - End of token (not in quotes)                           */
/*    '"'   - Start or end of token in quotes or embedded quote      */
/*    Other - Character in token                                     */
/*  end while                                                        */
/*                                                                   */
/*********************************************************************/
/*                                                                   */
/* Note: This code is taken from AMQSRESA.C                          */
/*                                                                   */
/*********************************************************************/
MQLONG GetNextToken(PMQCHAR *ppNameValueString,
                    PMQLONG  pRemainingLength,
                    PMQBYTE *ppToken,
                    ULONG   *pTokenLength)
{
  MQLONG       rc = OK;
  PMQCHAR      pCurrentChar = *ppNameValueString;
  Parser_State CurrentState = OutOfToken;   /* Finite state variable */

  /*******************************************************************/
  /* Set the output values to blank.                                 */
  /*******************************************************************/
  *ppToken = NULL;
  *pTokenLength = 0;

  /*******************************************************************/
  /* Step along the string one character at a time until either we   */
  /* reach the end of the string or the end of a token is reached.   */
  /* The processing of the string is based on a finite state machine */
  /* that has the following states:                                  */
  /*                                                                 */
  /*                                                                 */
  /* OutOfToken--->InToken----------------------->EndOfToken         */
  /*            |                             ^                      */
  /*            |                             |                      */
  /*             ->InQuotes--->EmbeddedQuote--                       */
  /*                  ^             |                                */
  /*                  |             |                                */
  /*                   -------------                                 */
  /*                                                                 */
  /*******************************************************************/
  while( (*pRemainingLength > 0)
       &&(CurrentState != EndOfToken)
       &&(rc == OK) )
  {
    switch( *pCurrentChar )
    {
      /***************************************************************/
      /* '\0' (null character)                                       */
      /* A null character is treated as a string terminater and any  */
      /* data in the string following the null is ignored.           */
      /* If we are either currently in a token (InToken) or the      */
      /* previous character was a '"' inside a quoted string         */
      /* (EmbeddedQuote) we can accept the null as the delimiter of  */
      /* the token and move to the EndOfToken State.                 */
      /***************************************************************/
      case '\0' :
        if( (CurrentState == EmbeddedQuote)
          ||(CurrentState == InToken) )
          CurrentState = EndOfToken;
        /*************************************************************/
        /* This is treated as the last character in the string so    */
        /* remaining length is currently this single character (1).  */
        /*************************************************************/
        *pRemainingLength = 1;
        break;
      /***************************************************************/
      /* ' ' (blank character)                                       */
      /* The blank is the token delimiter character (unless enclosed */
      /* in quotes), as therefore, is handle in the same way as the  */
      /* null character above.                                       */
      /***************************************************************/
      case ' ' :
        if( (CurrentState == EmbeddedQuote)
          ||(CurrentState == InToken) )
          CurrentState = EndOfToken;
        break;
      /***************************************************************/
      /* '"' (quote character)                                       */
      /***************************************************************/
      case '"' :
        switch( CurrentState )
        {
          /***********************************************************/
          /* If we are currently not yet in a token then the quote   */
          /* signifies that the token is to be enclosed in quotes    */
          /* and the token can be entered at this point (changes     */
          /* state to InQuotes).                                     */
          /***********************************************************/
          case OutOfToken :
            CurrentState = InQuotes;
            *ppToken = (PMQBYTE)pCurrentChar;
            break;
          /***********************************************************/
          /* If we are already inside a quote enclosed token  '"'    */
          /* character can either be the closing quote character     */
          /* (which must be followed by a delimiter character - see  */
          /* above) or it is an embedded quote (immediately          */
          /* followed by anther '"'). We change state to             */
          /* EmbeddedQuote and continue, testing the next character  */
          /* to determine the nature of this quote.                  */
          /***********************************************************/
          case InQuotes :
            CurrentState = EmbeddedQuote;
            break;
          /***********************************************************/
          /* The previous character is also a '"', therefore, the    */
          /* two quotes are an escaped quote embedded in the quoted  */
          /* token, we can change the state back to InQuotes and     */
          /* continue to find the end of the token.                  */
          /***********************************************************/
          case EmbeddedQuote :
            CurrentState = InQuotes;
            break;
          /***********************************************************/
          /* The only other state we could be in at this stage is    */
          /* InToken, it is invalid to have a '"' character embedded */
          /* in an un-quotes token.                                  */
          /***********************************************************/
          default :
            rc = NOT_OK;
            break;
        }
        break;
      /***************************************************************/
      /* Any other character                                         */
      /***************************************************************/
      default :
        switch( CurrentState )
        {
          /***********************************************************/
          /* Any non-delimiter character that follows either the     */
          /* start of the string or a delimiter character is taken   */
          /* as the start of a token, as this character cannot be a  */
          /* '"' (tested above) the token is not enclosed in quotes  */
          /* and the state is changed to InToken.                    */
          /***********************************************************/
          case OutOfToken :
            CurrentState = InToken;
            *ppToken = (PMQBYTE)pCurrentChar;
            break;
          /***********************************************************/
          /* An embedded quote ('"') found in a quoted token must be */
          /* followed either by another '"' (escaped quotes) or a    */
          /* delimiter character (' ' or '\0'), otherwise it is an   */
          /* error.                                                  */
          /***********************************************************/
          case EmbeddedQuote :
            rc = NOT_OK;
            break;
          default :
            break;
        }
        break;
    } /* end of switch( *pCurrentChar ) */
    /*****************************************************************/
    /* Step to the next character in the string and decrement the    */
    /* length of the remaining string.                               */
    /*****************************************************************/
    pCurrentChar++;
    (*pRemainingLength)--;
  } /* end of while */

  if( rc == OK )
  {
    /*****************************************************************/
    /* We have either reached the end of the string or the end of a  */
    /* token, check the state to determine which it is.              */
    /*****************************************************************/
    switch( CurrentState )
    {
      /***************************************************************/
      /* InToken, EmbeddedQuote or EndOfToken are all valid states   */
      /* to be in after finding a complete token. Calculate the      */
      /* length of the token before returning to the calling         */
      /* function. If we are in the EndOfToken state we will have    */
      /* moved on an extra character so take that into account.      */
      /***************************************************************/
      case InToken :
      case EmbeddedQuote :
        *pTokenLength = (PMQBYTE)pCurrentChar - *ppToken;
        break;
      case EndOfToken :
        *pTokenLength = (PMQBYTE)(pCurrentChar - 1) - *ppToken;
        break;
      /***************************************************************/
      /* To be in the OutOfToken state we could not have found even  */
      /* the start of a token, this could still be a valid state as  */
      /* the end of the string has been reached.                     */
      /***************************************************************/
      case OutOfToken :
        break;
      /***************************************************************/
      /* Any other state (InQuotes) are invalid states to end in, it */
      /* implies the start of a token has been found but not the end */
      /***************************************************************/
      default :
        rc = NOT_OK;
        break;
    }
  }

  /*******************************************************************/
  /* Move the string pointer to the point in the string that we have */
  /* got to so that the next call to this function will start at     */
  /* this point.                                                     */
  /*******************************************************************/
  *ppNameValueString = pCurrentChar;

  if( rc != OK )
    fprintf(pLogFile, "Invalid NameValueString\n");

  return(rc);
}
/*********************************************************************/
/* end of GetNextToken                                               */
/*********************************************************************/
